import json
import logging
import pickle
import time
from collections import defaultdict
from pathlib import Path
from typing import List

import dill
import numpy as np
import yaml
from omegaconf.dictconfig import DictConfig
from torch import randperm

log = logging.getLogger()
f_format_data = lambda data: [list(e) for e in zip(*data)]


def create_tmp_directory(storage_path, name):
    dir_path = Path(storage_path) / ("datasets/" + name + "_tmp")
    dir_path.mkdir(exist_ok=True)


def json_dump(obj, path):
    if isinstance(obj, np.ndarray):
        obj = obj.tolist()
    with open(path, "w") as f:
        json.dump(obj, f)


def json_load(path):
    with open(path) as f:
        obj = json.load(f)
    return obj


def pickle_dump(obj, path):
    with open(path, "wb") as f:
        pickle.dump(obj, f)


def pickle_load(path):
    with open(path, "rb") as f:
        obj = pickle.load(f)
    return obj


def dill_dump(obj, path):
    with open(path, "wb") as f:
        dill.dump(obj, f)


def dill_load(path):
    with open(path, "rb") as f:
        obj = dill.load(f)
    return obj


def read_yaml(file_path):
    with open(file_path) as f:
        data = yaml.load(f, yaml.Loader)
    return data


def tensor_to_numpy(tensor):
    return tensor.cpu().detach().numpy()


def create_time_dict(time_dict_path, model_name, mode="w"):
    init_dict = {model_name + "_fit": [], model_name + "_predict": []}
    with open(time_dict_path, mode) as f:
        json.dump(init_dict, f)


def add_new_model_to_time_dict(time_dict_path, new_model_name):
    with open(time_dict_path) as f:
        time_dict = json.load(f)
    time_dict.update({new_model_name + "_fit": [], new_model_name + "_predict": []})
    with open(time_dict_path, "w") as f:
        json.dump(time_dict, f)


def get_target_model_checkpoints(config, version="new", framework="transformers"):
    if version == "new":
        model_cfg = config.successor_model
    else:
        model_cfg = config.target_model
    checkpoints_path = Path(
        model_cfg.training.trainer_args.serialization_dir
    )
    model_name = "pytorch_model.pt" if framework == "flair" else "pytorch_model.bin"
    models_paths = [
        x / model_name
        for x in checkpoints_path.iterdir()
        if str(x).split("/")[-1].startswith("checkpoint")
    ]
    return models_paths


def get_num_chekpoints(config, name):
    # This part differs for the benchmark & for the real use
    if "post_processing" in config:
        if (not config.post_processing.tracin.use) or (name != "successor"):
            return 1
        return config.post_processing.tracin.num_model_checkpoints

    if ("tracin" not in config) or (not config.tracin.use) or (name != "target"):
        return 1
    return config.tracin.num_model_checkpoints


def get_time_dict_path(config):

    initial_time_dict = {
        "acquisition_fit": [],
        "acquisition_predict": [],
        "successor_fit": [],
        "successor_predict": [],
    }
    if "target_model" in config:
        initial_time_dict["target_fit"] = []
        initial_time_dict["target_predict"] = []

    # Create time dir if it is not created
    Path(config.cache_dir).mkdir(exist_ok=True)

    acquisition_name = config.acquisition_model.checkpoint.replace("/", "-")
    successor_name = (
        config.successor_model.checkpoint.replace("/", "-")
        if config.successor_model
        else acquisition_name
    )
    seed = config.seed

    if "target_model" not in config:
        file_name = f"time_dict_{acquisition_name}_{successor_name}_{seed}.json"
    else:
        target_name = config.target_model.checkpoint.replace("/", "-")
        file_name = (
            f"time_dict_{acquisition_name}_{successor_name}_{target_name}_{seed}.json"
        )

    time_dict_path = Path(config.cache_dir) / file_name
    json_dump(initial_time_dict, time_dict_path)

    return time_dict_path


def get_time_dict_path_full_data(config):

    initial_time_dict = {"model_fit": [], "model_predict": []}

    # Create time dir if it is not created
    Path(config.cache_dir).mkdir(exist_ok=True)

    model_name = config.model.checkpoint.replace("/", "-")
    seed = config.seed

    file_name = f"time_dict_{model_name}_{seed}.json"

    time_dict_path = Path(config.cache_dir) / file_name
    json_dump(initial_time_dict, time_dict_path)

    return time_dict_path


def get_config_to_update(dataset_configs_path, dataset_name, model_name):
    dataset_config = read_yaml(f"{dataset_configs_path}/{dataset_name}/dataset.yaml")
    model_config = read_yaml(
        f"{dataset_configs_path}/{dataset_name}/model_configs/{model_name}.yaml"
    )
    if dataset_config is not None:
        dataset_config.update(model_config or {})
    else:
        dataset_config = {}

    return dataset_config


def initialize_metrics_dict(
    model_type: str,  # cls, ner, ats, nmt, cv-cls
    work_dir: str or Path,
    model_name: str,
    framework: str = "transformers",
    evaluate_query: bool = False,
):
    """
    Create dict of metrics
    """
    metrics_dict = {}

    path_to_dump = Path(work_dir) / f"{model_name}_metrics.json"
    json_dump(metrics_dict, path_to_dump)
    if evaluate_query:
        path_to_dump = Path(work_dir) / f"{model_name}_evaluate_query_metrics.json"
        json_dump(metrics_dict, path_to_dump)


def get_metrics_dict(model_type):
    """
    Create dict of metrics
    """
    if model_type == "cls":
        keys = ["accuracy", "loss"]
    elif model_type == "ner":
        keys = [
            "accuracy",
            "accuracy3",
            "precision-overall",
            "recall-overall",
            "f1-measure-overall",
            "loss",
        ]
    else:  # Hence it is AutoNER
        keys = ["micro", "macro", "weighted"]

    metrics_dict = {}

    for phase in ["train", "test"]:
        for key in keys:
            metrics_dict[phase + "_" + key] = []

    return metrics_dict


def calculate_time_decorator(time_dict_path, step):
    def decorator(function):
        def wrapped(*args, **kwargs):
            start_time = time.time()

            result = function(*args, **kwargs)

            time_work = time.time() - start_time
            time_dict = json.load(time_dict_path)
            time_dict[step].append(time_work)
            json_dump(time_dict, time_dict_path)

            return result

        return wrapped

    return decorator


def log_config(log, conf, num_tabs=0):
    config = to_dict(conf)
    for key in config.keys():
        if isinstance(config[key], dict):
            log.info("\t" * num_tabs + key)
            log_config(log, config[key], num_tabs + 1)
        else:
            log.info("\t" * num_tabs + f"{key}: {config[key]}")


def to_dict(conf):
    config = dict(conf)
    if "hydra" in config.keys():
        del config["hydra"]
    for key in config.keys():
        if isinstance(config[key], DictConfig):
            config[key] = to_dict(config[key])
    config = dict(config)
    return config


def get_balanced_sample_indices(dataset, num_classes, n_per_digit=2) -> List[int]:
    """Given `target_classes` randomly sample `n_per_digit` for each of the `num_classes` classes."""
    permed_indices = randperm(len(dataset))

    if n_per_digit == 0:
        return []

    num_samples_by_class = defaultdict(int)
    initial_samples = []

    for i in range(len(permed_indices)):
        permed_index = int(permed_indices[i])
        label = dataset["label"][permed_index]
        index, target = permed_index, int(label)

        num_target_samples = num_samples_by_class[target]
        if num_target_samples == n_per_digit:
            continue

        initial_samples.append(index)
        num_samples_by_class[target] += 1

        if len(initial_samples) == num_classes * n_per_digit:
            break

    return initial_samples


def random_fixed_length_data_sampler(features):
    target_length = 5056
    indices = randperm(target_length + (-target_length % len(features)))
    indices = (indices[:target_length] % len(features)).tolist()
    new_features = features.select(indices)
    return new_features


class DictWithGetattr(dict):
    def __getattr__(self, item):
        return self.get(item, None)



import torch
import torch.nn as nn
from tokenizers import Tokenizer
from tokenizers.models import WordLevel
from tokenizers.normalizers import Lowercase
from tokenizers.pre_tokenizers import Whitespace
from transformers import PreTrainedTokenizerFast

from ..utils.general import DictWithGetattr


def init_tagger(model_cfg, embeddings, word2idx, num_labels, only_tokenizer=False):
    """Should return model and tokenizer instances.
    To add any custom model parameters just add them to model_cfg.
    """
    # here one could use custom function for embeddings
    # and override it, if necessary
    """
    if embeddings is None and model_cfg.embeddings_path is not None:
        embeddings, word2idx = load_embeddings(
            model_cfg.embeddings_path, model_cfg.embeddings_cache_dir
        )
    """
    # create tokenizer
    # in this example we use hf tokenizer, but one could also use any custom tokenizer as well
    tokenizer_model = WordLevel(word2idx, "[UNK]")
    tokenizer = Tokenizer(tokenizer_model)
    tokenizer.normalizer = Lowercase()
    tokenizer.pre_tokenizer = Whitespace()
    hf_tokenizer = PreTrainedTokenizerFast(
        tokenizer_object=tokenizer, pad_token="[PAD]", unk_token="[UNK]"
    )
    if only_tokenizer:
        return None, hf_tokenizer
    # build model
    model = BilstmTagger(
        model_config=model_cfg,
        num_labels=num_labels,
        pretrained_embeddings=embeddings,
        word2idx=word2idx,
    )
    return model, hf_tokenizer


class BilstmTagger(torch.nn.Module):
    def __init__(
        self,
        model_config: dict,
        num_labels: int,
        pretrained_embeddings: torch.Tensor = None,
        word2idx: dict = None,
    ):
        # model name shouldn't ends with Model
        super().__init__()
        self.model_config = model_config
        self.num_labels = num_labels
        self.pretrained_embeddings = pretrained_embeddings
        self.word2idx = word2idx
        # special key showing does model return embeddings in forward or not
        self.return_embeddings = False
        self.init_model()

    def init_model(self):
        if self.pretrained_embeddings is not None:
            self.vocab_size, self.embed_dim = self.pretrained_embeddings.shape
            # self.embed_dim - necessary argument for embeddings based strategies
            self.embedding = nn.Embedding.from_pretrained(
                self.pretrained_embeddings,
                padding_idx=self.vocab_size - 1,
                freeze=self.model_config.freeze_embedding,
            )
        else:
            self.embed_dim = self.model_config.embed_dim
            self.embedding = nn.Embedding(
                num_embeddings=self.model_config.vocab_size,
                embedding_dim=self.embed_dim,
                padding_idx=0,
                max_norm=5.0,
            )
        self.body = nn.LSTM(
            self.embed_dim, self.embed_dim // 4, 1, batch_first=True, bidirectional=True
        )
        self.head = nn.Sequential(
            nn.ReLU(),
            nn.Dropout(self.model_config.classifier_dropout),
            nn.Linear(self.embed_dim // 2, self.num_labels),
        )
        self.loss_fn = nn.CrossEntropyLoss()
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.to(self.device)

    def forward(self, input_ids, labels=None, output_hidden_states=False, **kwargs):
        x_embed = self.embedding(input_ids).float()
        # now shape is (bs, seq_len, embed_dim)
        x_body, _ = self.body(x_embed)
        # now shape is (bs, seq_len, embed_dim // 2)
        logits = self.head(x_body)
        # now shape is (bs, seq_len, num_labels)
        loss = None
        if labels is not None:
            loss = self.loss_fn(
                logits.reshape(-1, logits.shape[-1]), labels.reshape(-1)
            )
            results = {"loss": loss, "logits": logits}
        else:
            results = {"logits": logits}
        if output_hidden_states or self.return_embeddings:
            results["last_hidden_state"] = x_body
            results["hidden_states"] = [x_body]
        results = DictWithGetattr(results)
        return results

    def get_pre_classifier_dropout_activation(self):
        """Additional function - used for embeddings based AL strategies"""
        return None, self.head[-2], nn.ReLU()

    def save_pretrained(self, save_path):
        torch.save(self.state_dict(), save_path)

    def load_pretrained(self, load_path):
        self.load_state_dict(torch.load(load_path))



import logging
from pathlib import Path
from pandas import DataFrame

import numpy as np
from datasets import load_dataset, Dataset

from .preprocessing import (
    preprocess_img,
    _add_id_column_to_datasets,
    _use_test_subset,
    _use_train_subset,
    _filter_quantiles,
    _multiply_data,
)

log = logging.getLogger()


class HuggingFaceDatasetsReader:
    def __init__(
        self, *dataset_args, cache_dir=None, extra_key=None, use_auth_token=None
    ):
        self.dataset = load_dataset(
            *dataset_args, cache_dir=cache_dir, use_auth_token=use_auth_token
        )
        self.extra_key = extra_key

    def __call__(self, phase, text_name=None, label_name=None):
        dataset = self.dataset[phase]
        if self.extra_key is not None:
            dataset = Dataset.from_pandas(DataFrame(dataset[self.extra_key]))
        if text_name is not None and label_name is not None:
            dataset = dataset.remove_columns(
                [
                    x
                    for x in dataset.column_names
                    if x not in [text_name, label_name, "id"]
                ]
            )
        setattr(self, phase, dataset)
        return getattr(self, phase)


def load_huggingface_dataset(config, task, cache_dir=None):

    data_cache_dir = Path(cache_dir) / "data" if cache_dir is not None else None
    text_name = config.text_name
    label_name = config.label_name
    default_extra_key = "translation" if task == "nmt" else None
    extra_key = config.get("extra_key", default_extra_key)

    kwargs = {
        "cache_dir": data_cache_dir,
        "use_auth_token": config.get("use_auth_token", None),
        "extra_key": extra_key,
    }
    if config.get("revision_master"):
        kwargs["revision"] = "master"

    hfdreader = (
        HuggingFaceDatasetsReader(config.dataset_name, **kwargs)
        if isinstance(config.dataset_name, str)
        else HuggingFaceDatasetsReader(*list(config.dataset_name), **kwargs)
    )

    if config.get("multiply_data", None) is not None:
        hfdreader = _multiply_data(hfdreader, config.multiply_data)

    train_dataset = hfdreader("train", text_name, label_name)
    test_dataset = None
    if "validation" in hfdreader.dataset.keys():
        dev_dataset = hfdreader("validation", text_name, label_name)
        # Since on GLUE we do not have gold labels for test data
        if "test" in hfdreader.dataset.keys() and ("glue" not in config.dataset_name):
            test_dataset = hfdreader("test", text_name, label_name)
    else:
        dev_dataset = hfdreader("test", text_name, label_name)

    log.info(f"Loaded train size: {len(train_dataset)}")
    log.info(f"Loaded dev size: {len(dev_dataset)}")
    if test_dataset is None:
        log.info("Dev dataset coincides with test dataset")
    else:
        log.info(f"Loaded test size: {len(test_dataset)}")

    if config.labels_to_remove is not None:
        train_dataset = train_dataset.filter(
            lambda x: x[label_name] not in config.labels_to_remove
        )
        dev_dataset = dev_dataset.filter(
            lambda x: x[label_name] not in config.labels_to_remove
        )
        if test_dataset is not None:
            test_dataset = test_dataset.filter(
                lambda x: x[label_name] not in config.labels_to_remove
            )

    if task == "cls" or task == "cnn_cls":
        id2label = {
            i: val for i, val in enumerate(train_dataset.features[label_name].names)
        }
    elif task == "ner":
        id2label = {
            i: val
            for i, val in enumerate(train_dataset.features[label_name].feature.names)
        }
    elif task in ["ats", "nmt"]:
        id2label = None
    elif task == "cv_cls":
        id2label = {
            i: val for i, val in enumerate(train_dataset.features[label_name].names)
        }
        train_dataset = train_dataset.map(preprocess_img)
        dev_dataset = dev_dataset.map(preprocess_img)
        if test_dataset is not None:
            test_dataset = test_dataset.map(preprocess_img)
    else:
        raise NotImplementedError

    if getattr(config, "filter_quantiles", None) is not None:
        train_dataset = _filter_quantiles(
            train_dataset,
            config.filter_quantiles,
            cache_dir,
            text_name,
            config.tokenizer_name,
        )

    if getattr(config, "use_subset", None) is not None:
        train_dataset = _use_train_subset(
            train_dataset,
            config.use_subset,
            getattr(config, "seed", 42),
            task,
            label_name,
        )

    if ("id" not in train_dataset.column_names) and config.get("add_id_column", True):
        train_dataset, dev_dataset, test_dataset = _add_id_column_to_datasets(
            [train_dataset, dev_dataset, test_dataset]
        )

    if getattr(config, "use_test_subset", False):
        if test_dataset is None:
            test_dataset = dev_dataset
        test_dataset, subsample_idx = _use_test_subset(
            test_dataset,
            config.use_test_subset,
            getattr(config, "seed", 42),
            getattr(config, "subset_fixed_seed", False),
        )
        dev_dataset = dev_dataset.select(
            np.setdiff1d(np.arange(len(dev_dataset)), subsample_idx)
        )

    if test_dataset is None:
        test_dataset = dev_dataset

    return [train_dataset, dev_dataset, test_dataset, id2label]



import torch
from torch import nn
from tqdm import tqdm

DOUBLE_INFO = torch.finfo(torch.double)
JITTERS = [0, DOUBLE_INFO.tiny] + [10 ** exp for exp in range(-10, 0, 1)]


def centered_cov(x):
    return x.T @ x / (len(x) - 1)


def compute_density(log_logits, label_probs):
    return torch.sum((torch.exp(log_logits / 768) * label_probs), dim=1)


def get_gmm_log_probs(gaussians_model, embeddings):
    return gaussians_model.log_prob(embeddings[:, None, :])


def gmm_fit(embeddings, labels):
    num_classes = len(set(labels))
    with torch.no_grad():
        centroids = torch.stack(
            [torch.mean(embeddings[labels == c], dim=0) for c in range(num_classes)]
        )
        cov_matrix = torch.stack(
            [
                centered_cov(embeddings[labels == c] - centroids[c])
                for c in range(num_classes)
            ]
        )

    with torch.no_grad():
        for jitter_eps in JITTERS:
            try:
                jitter = jitter_eps * torch.eye(
                    cov_matrix.shape[1], device=cov_matrix.device,
                ).unsqueeze(0)
                gmm = torch.distributions.MultivariateNormal(
                    loc=centroids, covariance_matrix=(cov_matrix + jitter),
                )
            except RuntimeError as e:
                if "cholesky" in str(e):
                    continue
            except ValueError as e:
                if "The parameter covariance_matrix has invalid values" in str(e):
                    continue
            break

    return gmm, jitter_eps


def centered_cov_torch(x):
    n = x.shape[0]
    res = 1 / (n - 1) * x.t().mm(x)
    return res


def compute_density_(logits, class_probs):
    return torch.sum((torch.exp(logits) * class_probs), dim=1)


def class_probs(data_loader):
    num_classes = 10
    class_n = len(data_loader.dataset)
    class_count = torch.zeros(num_classes)
    for batch in data_loader:
        label = batch["labels"]
        class_count += torch.Tensor([torch.sum(label == c) for c in range(num_classes)])

    class_prob = class_count / class_n
    return class_prob


def gmm_evaluate(net, gaussians_model, loader, device, num_classes, storage_device):

    num_samples = len(loader.dataset)
    logits_N_C = torch.empty(
        (num_samples, num_classes), dtype=torch.float, device=storage_device
    )
    labels_N = torch.empty(num_samples, dtype=torch.int, device=storage_device)

    with torch.no_grad():
        start = 0
        for batch in tqdm(loader):
            data = batch["image"].to(device)  # torch.Size([128, 1, 28, 28])
            label = batch["labels"].to(device)

            logit_B_C = gmm_forward(
                net, gaussians_model, data, label
            )  # torch.Size([128, 10])

            end = start + len(data)
            logits_N_C[start:end].copy_(logit_B_C, non_blocking=True)
            labels_N[start:end].copy_(label, non_blocking=True)
            start = end

    return logits_N_C, labels_N


def gmm_forward(net, gaussians_model, data_B_X, label):  # torch.Size([128, 1, 28, 28])

    if isinstance(net, nn.DataParallel):
        features_B_Z = net.module(data_B_X, label)
        features_B_Z = net.module.feature
    else:
        features_B_Z = net(data_B_X, label)
        features_B_Z = net.feature  # torch.Size([128, 512])

    log_probs_B_Y = gaussians_model.log_prob(
        features_B_Z[:, None, :]
    )  # torch.Size([128, 10])

    return log_probs_B_Y


def gmm_fit_(embeddings, labels, num_classes=10):
    with torch.no_grad():
        classwise_mean_features = torch.stack(
            [torch.mean(embeddings[labels == c], dim=0) for c in range(num_classes)]
        )
        classwise_cov_features = torch.stack(
            [
                centered_cov_torch(embeddings[labels == c] - classwise_mean_features[c])
                for c in range(num_classes)
            ]
        )

    with torch.no_grad():
        for jitter_eps in JITTERS:
            try:
                jitter = jitter_eps * torch.eye(
                    classwise_cov_features.shape[1],
                    device=classwise_cov_features.device,
                ).unsqueeze(0)
                gmm = torch.distributions.MultivariateNormal(
                    loc=classwise_mean_features,
                    covariance_matrix=(classwise_cov_features + jitter),
                )
            except RuntimeError as e:
                if "cholesky" in str(e):
                    continue
            except ValueError as e:
                if (
                    "The parameter covariance_matrix has invalid values"
                    or "Expected parameter covariance_matrix" in str(e)
                ):
                    continue
            except ValueError as e:
                if "Expected parameter covariance_matrix" in str(e):
                    continue
            break

    return gmm, jitter_eps



import logging

from .wrapper_encoder_pytorch import PytorchEncoderWrapper
from ..transformers.wrapper_cls import WrapperCls
from ...models import PYTORCH_INIT_MODELS_DICT

log = logging.getLogger()


class PytorchClsWrapper(PytorchEncoderWrapper, WrapperCls):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_model_class(self):
        try:
            return PYTORCH_INIT_MODELS_DICT[self.model_config.checkpoint]["model_class"]
        except:
            model_class, _ = self.get_model_from_path()
            return model_class



import logging
import os
from pathlib import Path
from shutil import rmtree

import hydra
from omegaconf import OmegaConf

os.environ["WANDB_DISABLED"] = "true"

log = logging.getLogger()

OmegaConf.register_new_resolver(
    "to_string", lambda x: x.replace("/", "_").replace("-", "_")
)
OmegaConf.register_new_resolver(
    "get_patience_value", lambda dev_size: 1000 if dev_size == 0 else 3
)


@hydra.main(
    config_path=os.environ["HYDRA_CONFIG_PATH"],
    config_name=os.environ["HYDRA_CONFIG_NAME"],
)
def main(config):
    # Set the working directory
    auto_generated_dir = os.getcwd()
    os.chdir(hydra.utils.get_original_cwd())
    # Enable offline mode for HuggingFace libraries if necessary
    if config.offline_mode:
        os.environ["HF_DATASETS_OFFLINE"] = "1"
        os.environ["TRANSFORMERS_OFFLINE"] = "1"
        start_word = "Loading cached"
        end_word = "loaded"
    else:
        start_word = "Caching"
        end_word = "cached"
    # Cache everything
    cache_all(config, start_word, end_word)
    # Remove unused directory
    rmtree(auto_generated_dir)


def cache_model_and_tokenizer(task, model_name, cache_dir=None, end_word="cached"):

    from transformers import (
        AutoModelForSequenceClassification,
        AutoModelForTokenClassification,
        AutoModelForSeq2SeqLM,
        AutoTokenizer,
    )

    if cache_dir is not None:
        model_cache_dir = cache_dir / "model"
        tokenizer_cache_dir = cache_dir / "tokenizer"
    else:
        model_cache_dir = None
        tokenizer_cache_dir = None

    if task == "cls":
        AutoModelForSequenceClassification.from_pretrained(
            model_name, cache_dir=model_cache_dir
        )
    elif task == "ner":
        AutoModelForTokenClassification.from_pretrained(
            model_name, cache_dir=model_cache_dir
        )
    elif task in ["ats", "nmt"]:
        AutoModelForSeq2SeqLM.from_pretrained(model_name, cache_dir=model_cache_dir)
    log.info("=" * 10 + f" Model {model_name} successfully {end_word} " + "=" * 10)
    AutoTokenizer.from_pretrained(model_name, cache_dir=tokenizer_cache_dir)
    log.info("=" * 10 + f" Tokenizer {model_name} successfully {end_word} " + "=" * 10)


def cache_all(config, start_word="Caching", end_word="cached"):

    from datasets import load_dataset, load_metric

    cache_dir = Path(config.cache_dir) if config.cache_model_and_dataset else None
    # Cache models and tokenizers
    log.info("." * 10 + f" {start_word} model and tokenizer " + "." * 10)
    cache_model_and_tokenizer(
        config.acquisition_model.type,
        config.acquisition_model.checkpoint,
        cache_dir,
        end_word,
    )
    if getattr(config, "successor_model", None) is not None:
        cache_model_and_tokenizer(
            config.successor_model.type,
            config.successor_model.checkpoint,
            cache_dir,
            end_word,
        )
    if getattr(config, "target_model", None) is not None:
        cache_model_and_tokenizer(
            config.target_model.type,
            config.target_model.checkpoint,
            cache_dir,
            end_word,
        )
    # Cache dataset and dataset metrics
    metrics_cache_dir = cache_dir / "metrics"
    if config.data.path == "datasets":
        if cache_dir is not None:
            data_cache_dir = cache_dir / "data"
        else:
            data_cache_dir = None
            metrics_cache_dir = None
        log.info("." * 10 + f" {start_word} dataset " + "." * 10)
        if isinstance(config.data.dataset_name, str):
            load_dataset(config.data.dataset_name, cache_dir=data_cache_dir)
            log.info(
                "=" * 10
                + f" Dataset {config.data.dataset_name} successfully {end_word} "
                + "=" * 10
            )
            try:
                load_metric(config.data.dataset_name, cache_dir=metrics_cache_dir)
                log.info(
                    "=" * 10
                    + f" Dataset {config.data.dataset_name} metric successfully {end_word} "
                    + "=" * 10
                )
            except:
                pass
        else:
            load_dataset(*list(config.data.dataset_name), cache_dir=data_cache_dir)
            log.info(
                "=" * 10
                + f" Dataset {list(config.data.dataset_name)} successfully {end_word} "
                + "=" * 10
            )
            try:
                load_metric(
                    *list(config.data.dataset_name), cache_dir=metrics_cache_dir
                )
                log.info(
                    "=" * 10
                    + f" Dataset {list(config.data.dataset_name)} metric successfully {end_word} "
                    + "=" * 10
                )
            except:
                pass
    # Cache dataset metrics
    log.info("." * 10 + f" {start_word} metrics " + "." * 10)
    main_metric = (
        "accuracy"
        if config.task == "cls"
        else "seqeval"
        if config.task == "ner"
        else "rouge"
        if config.task == "ats"
        else "sacrebleu"
        if config.task == "nmt"
        else None
    )
    load_metric(main_metric, cache_dir=metrics_cache_dir)
    log.info(
        "=" * 10 + f" Main metric {main_metric} successfully {end_word} " + "=" * 10
    )
    eval_metrics = config.acquisition_model.training.trainer_args.eval_metrics
    if eval_metrics is not None:
        for metric in eval_metrics:
            load_metric(metric, cache_dir=metrics_cache_dir)
            log.info(
                "=" * 10
                + f" Additional metric {metric} successfully {end_word} "
                + "=" * 10
            )


if __name__ == "__main__":
    main()



import logging
from typing import Union, List, Tuple

from datasets.metric import Metric
from transformers import AutoModelForSequenceClassification, DataCollatorWithPadding

from .transformers_base_wrapper import TransformersBaseWrapper
from .wrapper_encoder import WrapperEncoder

log = logging.getLogger()


class WrapperCls(WrapperEncoder, TransformersBaseWrapper):

    """
    Returns a function that converts text to list of ids (of tokens)
    """

    @staticmethod
    def get_tokenize_function(
        tokenizer, text_name, label_name, test_mode: bool = False, **kwargs
    ):
        if test_mode:

            def tokenize_function(instances):
                encoding = tokenizer(instances[text_name], truncation=True)
                return encoding

        else:

            def tokenize_function(instances):
                encoding = tokenizer(instances[text_name], truncation=True)
                encoding["labels"] = instances[label_name]
                return encoding

        return tokenize_function

    @staticmethod
    def get_data_collator_class():
        return DataCollatorWithPadding

    @staticmethod
    def get_metric_name():
        return "test_accuracy"

    @staticmethod
    def get_datasets_metric_name():
        return "accuracy"

    @staticmethod
    def get_additional_datasets_metric_names():
        return ["f1"]

    @staticmethod
    def get_model_class():
        return AutoModelForSequenceClassification

    def get_kwargs_for_compute_metrics_fn(self):
        return {"num_labels": self.num_labels}

    @staticmethod
    def get_compute_metrics_function(
        metric,
        num_labels: int = 2,
        additional_metrics: Union[Tuple[Metric], List[Metric]] = tuple(),
    ):
        def compute_metrics(eval_preds):
            logits, labels, *inputs = eval_preds
            preds = logits.argmax(axis=-1)
            metrics_dict = metric.compute(predictions=preds, references=labels)
            for add_metric in additional_metrics:
                if add_metric.name == "f1" and num_labels > 2:
                    for average in ["micro", "macro", "weighted"]:
                        add_metric_dict = add_metric.compute(
                            predictions=preds, references=labels, average=average
                        )
                        metrics_dict.update({f"f1_{average}": add_metric_dict["f1"]})
                else:
                    add_metric_dict = add_metric.compute(
                        predictions=preds, references=labels
                    )
                    metrics_dict.update(add_metric_dict)
            return metrics_dict

        return compute_metrics



import logging
from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import take_idx
from ..utils.get_gradient_lengths import compute_egl_scores
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def egl(
    model, X_pool: Union[Dataset, TransformersDataset], n_instances: int, **egl_kwargs,
):
    kwargs = dict(
        # Necessary
        model=model,
        dataloader_or_data=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        tokenizer=model.tokenizer,
        task=model.task,
        text_name=model.data_config["text_name"],
        label_name=model.data_config["label_name"],
    )

    scores = compute_egl_scores(**kwargs)
    query_idx = np.argpartition(-scores, n_instances)[:n_instances]

    query = X_pool.select(query_idx)

    uncertainty_estimates = scores

    return query_idx, query, uncertainty_estimates



import logging

import numpy as np
from libact.base.dataset import ensure_sklearn_compat
from libact.base.interfaces import ProbabilisticModel, ContinuousModel

log = logging.getLogger("actleto")


class MultipleQueryStrategy:
    """The helper class for quering multiple instances from the unlabeled dataset

    Decorator for libact strategies that queries multiple instances from unlabeled
    dataset.

    """

    def __init__(self, impl, query_n=10):
        """
        Args:
            impl: the implementation of query strategy (libact compatible).
            query_n (int): number of unannotated examples to query.

        """
        self.impl = impl
        self.query_n = query_n

    def make_query(self):
        try:
            id_score_list = self.impl.make_query(return_score=True)[1]
            id_score_list.sort(key=lambda p: -p[1])
            return {sample for sample, _ in id_score_list[: self.query_n]}
        except TypeError:
            return {self.impl.make_query() for _ in range(self.query_n)}

    def update(self, *args, **kwargs):
        return self.impl.update(*args, **kwargs)


class MultipleTokenQueryStrategy:
    """The helper class for quering multiple sentences from the unlabeled dataset
    based on the number of tokens in the sentences.

    Decorator for libact strategies that queries multiple instances from unlabeled
    dataset.

    """

    def __init__(self, impl, n_tokens):
        """
        Args:
            impl: the implementation of query strategy (libact compatible).
            query_n (int): number of unannotated examples to query.

        """
        self.impl = impl
        self.n_tokens = n_tokens

    def make_query(self):
        try:
            id_score_list = self.impl.make_query(return_score=True)[1]
            id_score_list.sort(key=lambda p: -p[1])

            curr_index = 0
            curr_tokens = 0
            while curr_index < len(id_score_list) and curr_tokens < self.n_tokens:
                index = id_score_list[curr_index][0]
                curr_tokens += len(self.impl.dataset.data[index][0])
                curr_index += 1

            log.info(f"Sampled: {curr_tokens} tokens with {curr_index} sentences")

            return {sample for sample, _ in id_score_list[:curr_index]}
        except TypeError:
            return {self.impl.make_query() for _ in range(self.query_n)}

    def update(self, *args, **kwargs):
        return self.impl.update(*args, **kwargs)


class SklearnProbaAdapterWithUnlabeled(ProbabilisticModel):
    """The adaptor of sklearn models for libact strategies."""

    def __init__(self, clf):
        self._model = clf

    def train(self, dataset, *args, **kwargs):
        X, y = list(zip(*dataset.data))
        X = ensure_sklearn_compat(X)
        y = np.asarray(y)
        self._model.fit(X, y, *args, **kwargs)

    def predict(self, feature, *args, **kwargs):
        return self._model.predict(feature, *args, **kwargs)

    def score(self, testing_dataset, *args, **kwargs):
        X, y = list(zip(*testing_dataset.data))
        X = ensure_sklearn_compat(X)
        y = np.asarray(y)
        return self._model.score(X, y, *args, **kwargs)

    def predict_real(self, feature, *args, **kwargs):
        return self._model.predict_proba(feature, *args, **kwargs) * 2 - 1

    def predict_proba(self, feature, *args, **kwargs):
        return self._model.predict_proba(feature, *args, **kwargs)


class SklearnRealAdapter(ContinuousModel):
    """The adaptor of sklearn models for libact strategies."""

    def __init__(self, clf):
        self._model = clf

    def train(self, dataset, *args, **kwargs):
        self._model.fit(*(dataset.format_sklearn() + args), **kwargs)

    def predict(self, feature, *args, **kwargs):
        return self._model.predict(feature, *args, **kwargs)

    def score(self, testing_dataset, *args, **kwargs):
        return self._model.score(*(testing_dataset.format_sklearn() + args), **kwargs)

    def predict_real(self, feature, *args, **kwargs):
        return self._model.decision_function(feature, *args, **kwargs)



from typing import Union, Tuple
from datasets import Dataset
import numpy as np
from sklearn.feature_extraction.text import CountVectorizer

from ..utils.transformers_dataset import TransformersDataset


def ngram_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **strategy_kwargs,
) -> Tuple[np.ndarray, Union[Dataset, TransformersDataset], np.ndarray]:
    text_name = model.data_config["text_name"]
    ngram_range = strategy_kwargs.get("ngram_range", (1, 1))
    if isinstance(ngram_range, int):
        ngram_range = (ngram_range, ngram_range)
    truncate_texts = strategy_kwargs.get("truncate_texts", True)
    tokenizer = model.tokenizer
    tokenization_kwargs = {"truncation": False}
    if truncate_texts:
        tokenization_kwargs["truncation"] = True
    X_train_texts = tokenizer.batch_decode(
        tokenizer(X_train[text_name], **tokenization_kwargs)["input_ids"]
    )
    X_pool_texts = tokenizer.batch_decode(
        tokenizer(X_pool[text_name], **tokenization_kwargs)["input_ids"]
    )

    counter = CountVectorizer(ngram_range=ngram_range)
    unlabeled_ngramms = counter.fit_transform(X_pool_texts)
    labeled_ngramms = counter.transform(X_train_texts)

    ngrams_unl_freq = unlabeled_ngramms.sum(axis=0).A.ravel()
    ngrams_lab_freq = labeled_ngramms.sum(axis=0).A.ravel() + 1
    ngrams_scores = np.log(ngrams_unl_freq / ngrams_lab_freq)

    uncertainty_estimates = (
        unlabeled_ngramms.dot(ngrams_scores) / unlabeled_ngramms.sum(axis=1).A.ravel()
    )
    query_idx = np.argsort(-uncertainty_estimates)[:n_instances]
    query = X_pool.select(query_idx)
    return query_idx, query, uncertainty_estimates



from typing import Union, Tuple, List
import numpy as np
import logging
from datasets.metric import Metric

from .transformers_base_wrapper import TransformersBaseWrapper
from .wrapper_seq2seq import WrapperSeq2Seq
from ...utils.seq2seq_metrics import decode_all, calculate_bleu, calculate_additional_hf_metrics


log = logging.getLogger()


class WrapperNmt(WrapperSeq2Seq, TransformersBaseWrapper):
    @staticmethod
    def get_tokenize_function(
        tokenizer, text_name, label_name, test_mode: bool = False, **kwargs
    ):
        if test_mode:
            def tokenize_function(instances):
                encoded = tokenizer(instances[text_name], truncation=True)
                return encoded

        else:
            def tokenize_function(instances):
                encoded = tokenizer(instances[text_name], truncation=True)
                with tokenizer.as_target_tokenizer():
                    labels = tokenizer(instances[label_name], truncation=True)
                    encoded["labels"] = labels["input_ids"]
                return encoded

        return tokenize_function

    @staticmethod
    def get_metric_name():
        return "test_bleu"

    @staticmethod
    def get_datasets_metric_name():
        return "sacrebleu"

    @staticmethod
    def get_additional_datasets_metric_names():
        return []

    def get_kwargs_for_compute_metrics_fn(self):
        return {
            "tokenizer": self.tokenizer,
        }

    @staticmethod
    def get_compute_metrics_function(
        metric,
        tokenizer,
        additional_metrics: Union[Tuple[Metric], List[Metric]] = tuple(),
    ):
        assert tokenizer is not None, "Please provide tokenizer"

        def compute_metrics(eval_preds):
            decoded_preds, decoded_labels, decoded_texts = decode_all(
                eval_preds, tokenizer, use_additional_metrics=False
            )
            # Main metric
            result = calculate_bleu(decoded_preds, decoded_labels)
            # Add mean generated length
            prediction_lens = [
                np.count_nonzero(pred != tokenizer.pad_token_id)
                for pred in decoded_preds
            ]
            result["gen_len"] = np.mean(prediction_lens)
            # Sacrebleu
            sacrebleu_result = metric.compute(
                predictions=decoded_preds, references=[[lab] for lab in decoded_labels],
            )
            sacrebleu_result["sacrebleu"] = sacrebleu_result.pop("score")
            result.update(sacrebleu_result)

            add_result = calculate_additional_hf_metrics(additional_metrics, decoded_labels, decoded_preds)
            result.update(add_result)

            return result

        return compute_metrics



import types
from functools import partial

import torch


class BayesianModule(torch.nn.Module):
    """A module that we can sample multiple times from given a single input batch.

    To be efficient, the module allows for a part of the forward pass to be deterministic.
    """

    k = None

    def __init__(self):
        super().__init__()

    # Returns B x n x output
    def forward(self, input_B: torch.Tensor, k: int):
        BayesianModule.k = k

        mc_input_BK = BayesianModule.mc_tensor(input_B, k)
        mc_output_BK = self.mc_forward_impl(mc_input_BK)
        mc_output_B_K = BayesianModule.unflatten_tensor(mc_output_BK, k)
        return mc_output_B_K

    def mc_forward_impl(self, mc_input_BK: torch.Tensor):
        return mc_input_BK

    @staticmethod
    def unflatten_tensor(input: torch.Tensor, k: int):
        input = input.view([-1, k] + list(input.shape[1:]))
        return input

    @staticmethod
    def flatten_tensor(mc_input: torch.Tensor):
        return mc_input.flatten(0, 1)

    @staticmethod
    def mc_tensor(input: torch.tensor, k: int):
        mc_shape = [input.shape[0], k] + list(input.shape[1:])
        return input.unsqueeze(1).expand(mc_shape).flatten(0, 1)


class _ConsistentMCDropout(torch.nn.Module):
    def __init__(self, p=0.5):
        super().__init__()

        if p < 0 or p > 1:
            raise ValueError(
                "dropout probability has to be between 0 and 1, " "but got {}".format(p)
            )

        self.p = p
        # Sometimes one dropout layer can be used both for 3d tensors and 4d tensors (e.g. in XLNet)
        self.mask = {"3d": None, "4d": None}
        self.shape = None

    def extra_repr(self):
        return "p={}".format(self.p)

    def reset_mask(self):
        self.mask = {"3d": None, "4d": None}

    def train(self, mode=True):
        self.training = mode
        if not mode:
            self.reset_mask()

    def _create_mask(
        self,
        input: torch.Tensor = None,
        shape: torch.Size = None,
        share_across_tokens: bool = True,
        device: torch.device = None,
    ):
        assert ((shape is not None) and (device is not None)) or (
            input is not None
        ), "Either input or (shape and device) should be specified!"
        if input is not None:
            device = input.device
            if len(input.shape) == 3:
                if share_across_tokens:
                    mask_shape = [1, 1, 4096]
                else:
                    mask_shape = [1, 512, 4096]
            elif len(input.shape) == 4:
                if share_across_tokens:
                    mask_shape = [1, 16, 1, 1]
                else:
                    mask_shape = [1, 16, 512, 512]
            else:
                raise ValueError("Invalid shape of input!")
        else:
            mask_shape = [1] + list(shape)
        mask = torch.empty(mask_shape, dtype=torch.bool, device=device).bernoulli_(
            self.p
        )
        return mask

    def forward(self, input: torch.Tensor):
        if self.p == 0.0:
            return input

        self.shape = input.shape
        num_dim = len(input.shape)
        if self.training:
            # Create a new mask on each call and for each batch element.
            mask = self._create_mask(input)
        else:
            key = f"{num_dim}d"
            if self.mask[key] is None:
                # print('recreating mask', self)
                # Recreate mask.
                self.mask[key] = self._create_mask(input)

            mask = self.mask[key]

        truncated_mask = self.truncate_mask(mask)
        mc_output = input.masked_fill(truncated_mask, 0) / (1 - self.p)
        return mc_output

    def truncate_mask(self, mask):
        slicers = [slice(0, 1)] + [slice(0, x) for x in self.shape[1:]]
        truncated_mask = mask[slicers]
        return truncated_mask


def make_dropouts_consistent(model, share_across_tokens=True):
    for mod in model.modules():
        if type(mod).__name__ == "Dropout":
            for method in [
                "__init__",
                "extra_repr",
                "reset_mask",
                "train",
                "_create_mask",
                "forward",
                "truncate_mask",
            ]:
                setattr(
                    mod,
                    method,
                    types.MethodType(getattr(_ConsistentMCDropout, method), mod),
                )
            if not share_across_tokens:
                mod._create_mask = partial(mod._create_mask, share_across_tokens=False)
            mod.mask = None



from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import take_idx
from ..utils.transformers_dataset import TransformersDataset


def entropy(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    """
    Selects instances with the largest prediction entropy.
    https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.28.9963&rep=rep1&type=pdf
    """
    probas = model.predict_proba(X_pool)
    uncertainty_estimates = np.sum(-probas * np.log(probas), axis=1)
    argsort = np.argsort(-uncertainty_estimates)
    query_idx = argsort[:n_instances]
    query = X_pool.select(query_idx)

    return query_idx, query, uncertainty_estimates



import warnings
from pathlib import Path
from typing import List, Union, Tuple

from IPython.display import display
from pandas import DataFrame

from .extract_metrics import collect_data
from .plot_figure import plot_with_confidence_interval

warnings.filterwarnings("ignore")


def visualize_experiments(
    paths_to_experiments: List[
        Union[str, Path]
    ],  # list with paths to experiments (str or Path)
    num_queries: int = 16,  # Number of AL queries (=config.al.num_queries + 1)
    task: str = "cls",  # "cls" / "ner" / "ats" / "nmt"
    fill_missing: bool = False,  # whether to add an experiment with observations < `num_queries`
    prefix: str or None = "test",  # "test" or "eval" or None
    quantiles: Union[list, tuple] = None,  # quantiles to use if loading Tracin results
    log_errors: bool = False,  # whether to log the errors for unplotted experiments
    metric_name: str = None,  # if there is a specific metric you want to visualize (e.g. for abs-sum)
    metric_file_names: Union[
        List[str], Tuple[str]
    ] = None,  # if wanna use distinct files with metrics
    exclude_cache_from_dir: bool = True,  # if wanna exclude directory from search if it has "cache" in its name
    display_df_head: bool = False,  # whether we want to display the head of the `df` with experiments
    query_condition: str = None,  # query condition for observations from `df` which we want to plot
    columns_to_group_by=(
        "al_type",
        "strategy",
        "acquisition",
        "successor",
        "target",
    ),  # columns to groupby to separate different experiments
    save_name=None,  # path to save the plotted figure
    sleeve_type="conf",  # type of sleeve, either `conf` (confidence interval) or any other for ±std
    title="",  # plot title
    columns=None,  # how to name columns (otherwise names automatically)
    index=None,  # must have length equal to num_queries
    colors=None,  # colors for sleeves
    line_dash_sequence=None,  # how to plot different lines
    legend_font_size=20,  # size of the legend
    labels_font_size=15,  # size of the xlabel / ylabel text
    **kwargs  # other kwargs for px.line
):
    df = collect_data(
        paths_to_experiments,
        num_queries=num_queries,
        task=task,
        fill_missing=fill_missing,
        prefix=prefix,
        quantiles=quantiles,
        log_errors=log_errors,
        metric_name=metric_name,
        metric_file_names=metric_file_names,
        exclude_cache_from_dir=exclude_cache_from_dir,
    )
    if display_df_head:
        display(df.head())

    if query_condition is not None:
        df = df.query(query_condition)
    fig = plot_with_confidence_interval(
        df,
        columns_to_group_by=list(columns_to_group_by),
        save_name=save_name,
        metric_name=metric_name,
        task=task,
        sleeve_type=sleeve_type,
        title=title,  # plot title
        columns=columns,  # how to name columns (otherwise names automatically)
        index=index,  # must have length equal to num_queries
        colors=colors,  # colors for sleeves
        line_dash_sequence=line_dash_sequence,  # how to plot different lines
        legend_font_size=legend_font_size,  # size of the legend
        labels_font_size=labels_font_size,  # size of the xlabel / ylabel text
        **kwargs  # other kwargs for px.line
    )
    return fig


def plot_query(
    df: DataFrame,
    query_condition: str = None,  # query condition for observations from `df` which we want to plot
    **plot_with_confidence_interval_kwargs  # kwargs for `plot_with_confidence_interval`
):
    if query_condition is not None:
        df = df.query(query_condition)
    else:
        df = df.copy()
    fig = plot_with_confidence_interval(df, **plot_with_confidence_interval_kwargs)
    return fig



import logging
import time
from pathlib import Path
from typing import Union

import flair
import mlflow
import numpy as np
import torch
import transformers
from datasets import load_metric
from datasets.arrow_dataset import Dataset as ArrowDataset
from flair.data import Sentence, Corpus, Label
from torch.nn.functional import softmax
from tqdm import tqdm
from transformers import (
    set_seed,
    TrainerCallback,
)

from ...model_wrappers.flair.flair_trainer import FlairModelTrainer
from ...models.bilstm_crf import create_flair_model_tokenizer
from ...utils.general import DictWithGetattr
from ...utils.general import create_time_dict, json_dump, json_load
from ...utils.get_train_constants import get_train_constants
from ...utils.transformers_dataset import TransformersDataset

log = logging.getLogger()
transformers.logging.set_verbosity_info()

# import os
# os.environ["CUDA_VISIBLE_DEVICES"]="0"
# os.environ["CUDA_LAUNCH_BLOCKING"]="1"


class TrainingMetricsLogger(TrainerCallback):
    def on_evaluate(self, args, state, control, metrics, **kwargs):
        # Get train loss
        log.info(f"Epoch {metrics['epoch']}\nEvaluation Metrics:")
        log.info(
            f"Train Loss: {state.log_history[-2]['loss']}"
        )  # since [-1] is the dict with evaluation metrics
        [
            log.info(f"{k.replace('_', ' ').title()}: {v}")
            for k, v in metrics.items()
            if k != "epoch"
        ]


def iob_iobes(tags):
    """
    IOB -> IOBES
    """
    new_tags = []
    for i, tag in enumerate(tags):
        if tag == "O":
            new_tags.append(tag)
        elif tag.split("-")[0] == "B":
            if i + 1 != len(tags) and tags[i + 1].split("-")[0] == "I":
                new_tags.append(tag)
            else:
                new_tags.append(tag.replace("B-", "S-"))
        elif tag.split("-")[0] == "I":
            if i + 1 < len(tags) and tags[i + 1].split("-")[0] == "I":
                new_tags.append(tag)
            else:
                new_tags.append(tag.replace("I-", "E-"))
        else:
            raise Exception("Invalid IOB format!")
    return new_tags


def make_flair_sentences(X, y=None, idx2tag=None, tag_type=None):
    iobes = False
    sentences = [Sentence(tokens) for tokens in X]
    if y is not None:
        assert tag_type, "Tag type is required if tags (y) are defined"
        for sentence, tags in zip(sentences, y):
            if iobes:
                intermediate_tags = [idx2tag[tag] for tag in tags]
                intermediate_tags = iob_iobes(intermediate_tags)
                for (token, tag) in zip(sentence.tokens, intermediate_tags):
                    token.add_tag_label(tag_type, Label(tag))
            else:
                for (token, tag) in zip(sentence.tokens, tags):
                    token.add_tag_label(tag_type, Label(idx2tag[tag]))
    return sentences


def transform_idx2tag(data, tag_type=None):
    unique_tags = {"O"}
    tag_to_idx = {"O": 0}
    for sentence in data:
        ner_tags = []
        for token in sentence.tokens:
            ner_tags.append(token.get_tag(tag_type).value)
        unique_tags.update(ner_tags)
        if unique_tags != set(tag_to_idx.keys()):
            for key in unique_tags:
                if key not in tag_to_idx:
                    tag_to_idx[key] = len(tag_to_idx)
    idx_to_tag = {i: tag for tag, i in tag_to_idx.items()}
    return idx_to_tag


def transformers2flair_dataset(
    data: TransformersDataset,
    idx2tag: dict,
    text_name: str = "tokens",
    label_name: str = "ner_tags",
):
    """Transform transformers dataset to flair corpus.
    Implemented because we use flair alongside transformers.
    """
    if isinstance(data, ArrowDataset):
        instances = data
    else:
        instances = data.instances
        if idx2tag is None:
            idx2tag = data.id2label
    if instances[0].get(label_name) is not None:
        corpus = make_flair_sentences(
            [instance[text_name] for instance in instances],
            [instance[label_name] for instance in instances],
            idx2tag,
            label_name,
        )
    else:
        corpus = make_flair_sentences([instance[text_name] for instance in instances])
    return corpus


class FlairModelWrapper:  # TODO: add interface
    def __init__(
        self,
        model_config,
        checkpoint,
        task: str,
        model=None,
        tokenizer=None,
        name: str = "acquisition",
        default_data_config=None,
        seed: int = 42,
        dev_data_kwargs=None,
        trainer_kwargs=None,
        batch_size_kwargs=None,
        optimizer_kwargs=None,
        scheduler_kwargs=None,
        time_dict_path: Path or str = None,
        cache_dir: Path or str = None,
        cache_model: bool = False,
        num_checkpoints_to_save: int = 1,
        num_labels: int = 9,
        id2label=None,
        *args,
        **kwargs,
    ):
        self.checkpoint = checkpoint
        self.task = task
        self.name = name

        self.model = model
        self.model_config = model_config
        if tokenizer is not None:
            self.tokenizer = tokenizer
        else:
            self.set_tokenizer()

        self.num_labels = num_labels

        if default_data_config is None:
            default_data_config = {
                "dataset_name": None,
                "text_name": "text",
                "label_name": "label",
            }
        self.data_config = dict(default_data_config)
        self.data_config["idx_to_tag"] = id2label
        # TODO: make as an argument
        self.data_config["display_tag_stats"] = "f1"

        self._optimizer_kwargs = optimizer_kwargs
        self._trainer_kwargs = trainer_kwargs
        self._scheduler_kwargs = scheduler_kwargs
        self._batch_size_kwargs = batch_size_kwargs

        self.dev_data = dev_data_kwargs.pop("dev_data")
        self._dev_kwargs = DictWithGetattr(dev_data_kwargs)
        if self.dev_data is not None:
            # now it's flair Corpus
            self.tokenized_dev_data = transformers2flair_dataset(
                self.dev_data,
                self.data_config["idx_to_tag"],
                self.data_config["text_name"],
                self.data_config["label_name"],
            )

        self.seed = seed

        if self._trainer_kwargs["serialization_dir"] is None:
            self._trainer_kwargs["serialization_dir"] = "./output/"
        self._num_checkpoints_to_save = num_checkpoints_to_save
        if time_dict_path is None:
            time_dict_path = "time_dict.json"
            create_time_dict(time_dict_path, name)
        self.time_dict_path = time_dict_path
        self.cache_dir = (
            Path(cache_dir) if cache_dir is not None else Path(f"workdir/cache_{seed}")
        )
        self.cache_dir.mkdir(exist_ok=True)
        self.cache_model = cache_model

        self.device_name = "cuda"
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        flair.device = self.device
        if model is not None:
            self.model.to(flair.device)
        # Path(serialization_dir).mkdir(exist_ok=True)

    def set_tokenizer(self):
        self.tokenizer = None

    def fit(self, train_data, from_scratch=True, is_tokenized=False, data_config=None):

        # TODO: find a better way
        if self._dev_kwargs["size"] > 0:
            splitted = train_data.train_test_split(
                train_size=self._dev_kwargs[
                    "size"
                ],  # this is done to have the sample in reversed way
                shuffle=self._dev_kwargs["shuffle_dev"],
                seed=self.seed,
            )
            # this is done to have the sample in reversed way
            train_data = splitted["test"]
            dev_data = splitted["train"]

            if not is_tokenized:
                data_config = (
                    data_config if data_config is not None else self.data_config
                )
                dev_data = transformers2flair_dataset(
                    dev_data,
                    self.data_config["idx_to_tag"],
                    self.data_config["text_name"],
                    self.data_config["label_name"],
                )
        else:
            dev_data = self.dev_data
            if dev_data is not None:
                if data_config is None:
                    dev_data = self.tokenized_dev_data
                elif dev_data.tokenized_data is not None:
                    dev_data = dev_data.tokenized_data
                else:
                    dev_data = transformers2flair_dataset(
                        dev_data,
                        self.data_config["idx_to_tag"],
                        self.data_config["text_name"],
                        self.data_config["label_name"],
                    )

        log.info(f"Training dataset size: {len(train_data)}")
        log.info(
            f"Validation dataset size: {len(dev_data) if dev_data is not None else 0}"
        )

        (
            batch_size,
            num_epochs,
            steps_per_epoch,
            scheduler_warmup_steps,
        ) = get_train_constants(
            len(train_data),
            self._trainer_kwargs.num_epochs,
            self._batch_size_kwargs.batch_size,
            self._batch_size_kwargs.adjust_batch_size,
            self._batch_size_kwargs.adjust_num_epochs,
            self._batch_size_kwargs.min_num_gradient_steps,
            self._scheduler_kwargs.warmup_steps_factor,
            self._batch_size_kwargs.min_batch_size,
        )

        data_config = data_config if data_config is not None else self.data_config
        if not is_tokenized:
            train_data = transformers2flair_dataset(
                train_data,
                self.data_config["idx_to_tag"],
                self.data_config["text_name"],
                self.data_config["label_name"],
            )
        if from_scratch:
            self.model, self.tokenizer = create_flair_model_tokenizer(
                self.model_config,
                self.data_config["idx_to_tag"],
                self.data_config["label_name"],
                self.seed,
            )
            self.model.to(self.device)

        corpus = Corpus(train=train_data, dev=dev_data, test=[])
        # Here we use base SGD optimizer, as in prev code
        # set seed
        # torch.backends.cudnn.enabled = False
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        set_seed(self.seed)
        trainer = FlairModelTrainer(
            self.model,
            corpus,
            device=self.device_name,
            golden_label_type=self.data_config["label_name"],
        )
        start_time = time.time()
        results = trainer.train(
            base_path=self.model_config.training.trainer_args.serialization_dir,
            learning_rate=self._optimizer_kwargs.lr,
            mini_batch_size=batch_size,
            mini_batch_chunk_size=None,
            max_epochs=num_epochs,
            anneal_factor=self._scheduler_kwargs.factor,
            patience=self._scheduler_kwargs.patience,
            min_learning_rate=self._scheduler_kwargs.min_lr,
            train_with_dev=False,
            monitor_train=False,
            monitor_test=False,
            embeddings_storage_mode="cuda",
            checkpoint=True,
            save_final_model=True,
            anneal_with_restarts=True,
            anneal_with_prestarts=False,
            batch_growth_annealing=False,
            shuffle=True,
            param_selection_mode=False,
            num_workers=0,
            sampler=None,
            use_amp=False,
            amp_opt_level="O1",
            eval_on_train_fraction=0.0,
            eval_on_train_shuffle=False,
        )
        self._calculate_time(start_time, phase="fit")

        mlflow.end_run()
        self.trainer = trainer

    def get_predictions(
        self,
        data,
        is_tokenized=False,
        data_config=None,
        calculate_time=False,
        save_length_array=False,
        only_metrics=False,
    ):

        if data_config is None:
            data_config = self.data_config

        text_name = data_config["text_name"]
        label_name = data_config["label_name"]
        dataset_name = data_config["dataset_name"]
        metrics_cache_dir = (
            Path(self.cache_dir) / "metrics" if self.cache_dir is not None else None
        )

        labels = data[label_name]
        if not is_tokenized:
            data = transformers2flair_dataset(
                data, self.data_config["idx_to_tag"], text_name, label_name,
            )

        start_time = time.time()

        result = self._model_predict_loop(data)
        loss, logits, preds = result["loss"], result["logits"], result["predictions"]
        # pad logits and make np array

        compute_metrics_fn = self._return_compute_metrics_func(
            self.task,
            eval_metrics=None,
            idx_to_tag=self.data_config["idx_to_tag"],
            num_labels=self.num_labels,
            dataset_name=dataset_name,
            metrics_cache_dir=metrics_cache_dir,
            preprocessed_logits=False,
        )

        metrics_dict = compute_metrics_fn([logits, labels])
        metrics_dict["loss"] = loss
        metrics_dict = {k: v for k, v in metrics_dict.items()}

        # for normal saving
        metrics_dict = {key: float(value) for key, value in metrics_dict.items()}
        # add f1_score from model
        f1_flair = self.model.evaluate(
            data,
            self.data_config["label_name"],
            embedding_storage_mode="cuda",
            mini_batch_size=self._batch_size_kwargs["eval_batch_size"],
        )
        metrics_dict["f1_flair"] = float(f1_flair.main_score)

        if only_metrics:
            list_logits = []
        else:
            list_logits = logits.tolist()
        predictions = DictWithGetattr(
            {"predictions": list_logits, "metrics": metrics_dict}
        )

        if calculate_time:
            self._calculate_time(start_time, phase="predict")

        if save_length_array:
            self._tmp_length_array = [len(x) for x in labels]
        return predictions

    def predict_logits(
        self, data, is_tokenized=False, data_config=None, save_length_array=False
    ):
        predictions = self.get_predictions(
            data,
            is_tokenized,
            data_config,
            calculate_time=True,
            save_length_array=save_length_array,
        )
        return predictions.predictions

    def predict_proba(
        self,
        data,
        is_tokenized=False,
        data_config=None,
        to_numpy=True,
        remove_padding=False,
    ):
        logits = self.predict_logits(
            data,
            is_tokenized=is_tokenized,
            data_config=data_config,
            save_length_array=remove_padding,
        )
        probas = softmax(torch.Tensor(logits).to(self.device), dim=-1)

        if self.task == "ner" and remove_padding:
            probas = self._remove_padding(probas, self._tmp_length_array)
        elif to_numpy:
            return probas.cpu().detach().numpy()
        return probas

    def evaluate(self, data, is_tokenized=False, data_config=None, *args, **kwargs):

        predictions = self.get_predictions(
            data, is_tokenized, data_config, calculate_time=False, only_metrics=True,
        )
        return dict(predictions.metrics)

    def tokenize_data(
        self,
        tokenizer,
        data,
        task="cls",
        text_name="text",
        label_name="labels",
        **kwargs,
    ):
        # For flair model we simply transform dataset into flair corpus
        # Attention! This function supported only for compatibility with tracin
        if task == "ner":
            tokenized_data = transformers2flair_dataset(
                data, self.data_config["idx_to_tag"], text_name, label_name,
            )
            return tokenized_data
        else:
            raise NotImplementedError

    @staticmethod
    def _get_tokenize_fn_for_cls(
        tokenizer, text_name="text", label_name="labels", **kwargs
    ):
        if label_name == "labels":

            def tokenize_function(instance):
                encoded = tokenizer(instance[text_name], truncation=True, **kwargs)
                return encoded

        else:

            def tokenize_function(instance):
                encoded = tokenizer(instance[text_name], truncation=True, **kwargs)
                encoded["labels"] = instance[label_name]
                return encoded

        return tokenize_function

    @staticmethod
    def _get_tokenize_fn_for_ner(
        tokenizer, tokens_name="tokens", tags_name="ner_tags", **kwargs
    ):
        def tokenize_function(instance):
            encoded = tokenizer(
                instance[tokens_name],
                is_split_into_words=True,
                return_offsets_mapping=True,
                truncation=True,
                **kwargs,
            )
            ### Use offset mapping to restore the positions of original tokens and adapt ner tags to them
            # Initialize with zeros ('O' tag)
            encoded["labels"] = np.zeros(len(encoded["input_ids"]), dtype="int") - 100
            arr_offset = np.array(encoded["offset_mapping"])
            # Substitute values corresponding to first bpe-tokens with the tags of the tokens
            encoded["labels"][
                (arr_offset[:, 0] == 0) & (arr_offset[:, 1] != 0)
            ] = instance[tags_name]
            # Remove redundant items
            del encoded["offset_mapping"]
            return encoded

        return tokenize_function

    @staticmethod
    def _return_compute_metrics_func(
        task,
        eval_metrics=None,
        num_labels=2,
        dataset_name=None,
        idx_to_tag=None,
        display_tag_stats=None,
        metrics_cache_dir=None,
        preprocessed_logits=False,
    ):

        load_func = lambda task: load_metric(
            "accuracy" if task == "cls" else "seqeval", cache_dir=metrics_cache_dir
        )
        if dataset_name is not None:
            # since metrics are not implemented for all the datasets
            try:
                metric = (
                    load_metric(dataset_name, cache_dir=metrics_cache_dir)
                    if isinstance(dataset_name, str)
                    else load_metric(*dataset_name, cache_dir=metrics_cache_dir)
                )
            except:
                metric = load_func(task)
        elif dataset_name is not None:
            metric = load_metric(*dataset_name, cache_dir=metrics_cache_dir)
        else:
            metric = load_func(task)

        if eval_metrics is not None:
            additional_metrics = [
                load_metric(add_metric, cache_dir=metrics_cache_dir)
                for add_metric in eval_metrics
            ]
        elif task == "cls":
            additional_metrics = [load_metric("f1", cache_dir=metrics_cache_dir)]
            if metric.name != "accuracy":
                additional_metrics.append(
                    load_metric("accuracy", cache_dir=metrics_cache_dir)
                )
        elif task == "ner":
            additional_metrics = (
                [load_metric("seqeval", cache_dir=metrics_cache_dir)]
                if metric.name != "seqeval"
                else []
            )
        else:
            raise NotImplementedError

        if task == "cls":
            return FlairModelWrapper._return_compute_metrics_func_for_cls(
                metric, additional_metrics, num_labels
            )
        elif task == "ner":
            return FlairModelWrapper._return_compute_metrics_func_for_ner(
                metric,
                idx_to_tag,
                additional_metrics,
                num_labels,
                display_tag_stats,
                preprocessed_logits,
            )

    @staticmethod
    def _return_compute_metrics_func_for_cls(
        metric, additional_metrics: Union[tuple, list] = tuple(), num_labels=2
    ):
        def compute_metrics(eval_preds):
            logits, labels = eval_preds
            preds = logits.argmax(axis=-1)
            metrics_dict = metric.compute(predictions=preds, references=labels)
            for add_metric in additional_metrics:
                if add_metric.name == "f1" and num_labels > 2:
                    for average in ["micro", "macro", "weighted"]:
                        add_metric_dict = add_metric.compute(
                            predictions=preds, references=labels, average=average
                        )
                        metrics_dict.update({f"f1_{average}": add_metric_dict["f1"]})
                else:
                    add_metric_dict = add_metric.compute(
                        predictions=preds, references=labels
                    )
                    metrics_dict.update(add_metric_dict)

            return metrics_dict

        return compute_metrics

    @staticmethod
    def _return_compute_metrics_func_for_ner(
        metric,
        idx_to_tag,
        additional_metrics=tuple(),
        num_labels=2,
        display_tag_stats="f1",
        preprocessed_logits=False,
    ):
        def compute_metrics(eval_preds):
            logits_ids, labels_ids = eval_preds
            if not (preprocessed_logits):
                preds_ids = logits_ids.argmax(axis=-1)
                # Convert ids to tags
                length_array = FlairModelWrapper._get_length_array_from_label_ids(
                    labels_ids
                )
                labels = FlairModelWrapper._transform_idx_to_tag(
                    labels_ids, idx_to_tag, length_array
                )
                preds = FlairModelWrapper._transform_idx_to_tag(
                    preds_ids, idx_to_tag, length_array
                )
            else:
                length_array = FlairModelWrapper._get_length_array_from_label_ids(
                    labels_ids
                )
                labels = FlairModelWrapper._transform_idx_to_tag(
                    labels_ids, idx_to_tag, length_array
                )
                preds = logits_ids
            # Compute metrics
            metrics_dict = metric.compute(predictions=preds, references=labels)

            for add_metric in additional_metrics:
                if add_metric.name == "f1" and num_labels > 2:
                    for average in ["micro", "macro", "weighted"]:
                        add_metric_dict = add_metric.compute(
                            predictions=preds, references=labels, average=average
                        )
                        metrics_dict.update({f"f1_{average}": add_metric_dict["f1"]})
                else:
                    add_metric_dict = add_metric.compute(
                        predictions=preds, references=labels
                    )
                    metrics_dict.update(add_metric_dict)
            # Individual entity scores are presented as dicts now. Need to "straighten" them
            output_dict = FlairModelWrapper._construct_metrics_dict_for_ner(
                metrics_dict, display_tag_stats
            )
            return output_dict

        return compute_metrics

    def _token_probas_to_idx(self, sentences, label_name):
        """Get token probas from token.get_tags_proba_dist and
        return its in logits format. START and STOP tags counts as O tag.
        """
        extended_tag2idx = {
            value: key for key, value in self.data_config["idx_to_tag"].items()
        }
        extended_tag2idx["<START>"] = extended_tag2idx["O"]
        extended_tag2idx["<STOP>"] = extended_tag2idx["O"]
        all_probas = []
        all_preds = []
        for sentence in tqdm(sentences):
            probas = np.zeros((len(sentence), len(self.data_config["idx_to_tag"])))
            preds = [token.get_tag(label_name).value for token in sentence]
            all_preds.append(preds)
            for idx, el in enumerate(sentence):
                token = el.get_tags_proba_dist(label_name)
                for label in token:
                    class_idx = extended_tag2idx[label.value]
                    probas[idx][class_idx] += label.score
            all_probas.append(probas)
        return all_probas, all_preds

    def _model_predict_loop(self, data):
        # do the same, but without losses
        predictions = self.model.predict(
            data,
            mini_batch_size=self._batch_size_kwargs.eval_batch_size,
            all_tag_prob=True,
            label_name=self.data_config["label_name"] + "_pred",
            return_loss=True,
        )
        loss = predictions[0].detach().cpu().item() / predictions[1]
        logits, preds = self._token_probas_to_idx(
            data, label_name=self.data_config["label_name"] + "_pred"
        )
        # pad logits to make np array
        max_len = np.max([len(el) for el in logits])
        padded_logits = np.array(
            [np.pad(logit, ((0, max_len - logit.shape[0]), (0, 0))) for logit in logits]
        )
        return {"loss": loss, "logits": padded_logits, "predictions": preds}

    @staticmethod
    def _transform_idx_to_tag(array_idx, idx_to_tag, length_array):
        array_tag = []
        for boundary, instance_idx in zip(length_array, array_idx):
            instance_tag = []
            # TODO: There was 1:boundary - why?
            for idx in instance_idx[:boundary]:
                instance_tag.append(idx_to_tag[idx])
            array_tag.append(instance_tag)
        return array_tag

    @staticmethod
    def _get_length_array_from_label_ids(label_ids, padding_idx=-100):
        length_array = []
        for instance_ids in label_ids:
            inst_length = (
                np.argmax(instance_ids[1:] == padding_idx) + 1
                if instance_ids[-1] == padding_idx
                else len(instance_ids)
            )
            length_array.append(inst_length)
        return length_array

    @staticmethod
    def _construct_metrics_dict_for_ner(
        metrics_dict, display_tag_stats: str or bool = "f1"
    ):
        output_dict = {}
        for key, val in metrics_dict.items():
            if isinstance(val, dict):
                if display_tag_stats is False or display_tag_stats is None:
                    continue
                elif display_tag_stats == "f1":
                    output_dict[key + "_f1"] = val["f1"]
                else:
                    for key_2, val_2 in val.items():
                        output_dict[key + "_" + key_2] = val_2
            else:
                output_dict[key] = val

        return output_dict

    @staticmethod
    def _remove_padding(predictions, length_array):
        preds = []
        for i, pred in enumerate(predictions):
            boundary = (
                length_array[i] - 1
            )  # Since we do not need the prediction for "[SEP]"
            pred_without_padding = pred[1:boundary].cpu().detach().numpy()
            preds.append(pred_without_padding)

        return np.array(preds)

    def _calculate_time(self, start_time, phase="fit"):
        time_work = time.time() - start_time
        time_dict = json_load(self.time_dict_path)
        time_dict[self.name + f"_{phase}"].append(time_work)
        json_dump(time_dict, self.time_dict_path)
        log.info(f"Done with the model {phase}.")



import torch
from tqdm import tqdm

JITTERS = [10 ** exp for exp in range(-15, 0, 1)]


def compute_inv_covariance_v2(centroids, train_features, train_labels, jitters=None):
    jitter = 0
    jitter_eps = None
    if jitters is None:
        jitters = JITTERS

    cov = torch.zeros(
        len(centroids),
        centroids.shape[1],
        centroids.shape[1],
        device=centroids.device,
        dtype=torch.float32,
    ).float()
    for c, mu_c in tqdm(enumerate(centroids)):
        features_of_class_c = train_features[train_labels == c]
        for x in features_of_class_c:
            d = (x - mu_c).unsqueeze(1)
            cov[c] += d @ d.T
        cov[c] /= len(features_of_class_c)

    for i, jitter_eps in enumerate(jitters):
        jitter = jitter_eps * torch.eye(cov.shape[1], device=cov.device)
        cov_update = cov + jitter
        eigenvalues = torch.symeig(cov_update).eigenvalues
        if (eigenvalues >= 0).all():
            break
    cov = cov + jitter
    cov_inv = torch.empty(cov.shape, dtype=torch.float32, device=cov.device)
    for i in range(len(centroids)):
        cov_inv[i].copy_(torch.inverse(cov[i].to(torch.float64)).float())

    return cov_inv, jitter_eps


def mahalanobis_distance_with_known_centroids_sigma_inv_v2(
    centroids, centroids_mask, sigmas_inv, eval_features
):
    device = eval_features.device
    diffs = eval_features.unsqueeze(1) - centroids.unsqueeze(
        0
    )  # bs (b), num_labels (c / l), dim (d / a)
    dists = (
        torch.sqrt(torch.einsum("bld,lda,bla->bl", diffs, sigmas_inv, diffs))
        .squeeze(-1)
        .squeeze(-1)
        .cpu()
    )
    dists = dists.masked_fill_(centroids_mask, float("inf")).to(device)
    return dists



import logging
from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import (
    take_idx,
    choose_cm_samples,
)
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def cluster_margin(
    model, X_pool: Union[Dataset, TransformersDataset], n_instances: int, **kwargs
):
    """
    Select a diverse set of examples on which the model is least confident. The confidence of a model
    on an example is a difference between the largest two predicted class probabilities.In order to ensure diversity
    among the least confident examples, they are clustered using HAC with average-linkage.
    https://arxiv.org/pdf/2107.14263.pdf
    """
    instances_multiplier = kwargs.get("instances_multiplier", 1.25)
    random_query = kwargs.get("random_query", True)
    probas = model.predict_proba(X_pool)
    # reuse code from margin sampling
    # To get second max probas, need to sort the array since `.sort` modifies the array
    probas.sort(axis=1)
    max_probas = probas[:, -1]
    second_max_probas = probas[:, -2]
    uncertainty_estimates = 1 + second_max_probas - max_probas
    argsort = np.argsort(-uncertainty_estimates)
    # we have to choose n_instances with round-robin algorithm, so for this we
    # firstly choose 2 * n_instances
    query_idx = argsort[: int(instances_multiplier * n_instances)]
    cluster_labels = np.array(X_pool.clusters)[query_idx]
    # X_pool either transformer dataset or np array
    query_idx, samples_idx = choose_cm_samples(
        cluster_labels, query_idx, n_instances, random_query
    )
    query = take_idx(X_pool, query_idx)
    return query_idx, query, uncertainty_estimates



# %%
import torch
import torch.nn as nn
import traceback
from transformers import BartTokenizer, BartForConditionalGeneration
from typing import List
import numpy as np


class BARTScorer:
    def __init__(self, device='cuda:0', max_length=1024, checkpoint='facebook/bart-large-cnn'):
        # Set up model
        self.device = device
        self.max_length = max_length
        self.tokenizer = BartTokenizer.from_pretrained(checkpoint)
        self.model = BartForConditionalGeneration.from_pretrained(checkpoint)
        self.model.eval()
        self.model.to(device)

        # Set up loss
        self.loss_fct = nn.NLLLoss(reduction='none', ignore_index=self.model.config.pad_token_id)
        self.lsm = nn.LogSoftmax(dim=1)

    def load(self, path=None):
        """ Load model from paraphrase finetuning """
        if path is None:
            path = 'models/bart.pth'
        self.model.load_state_dict(torch.load(path, map_location=self.device))

    def score(self, srcs, tgts, batch_size=4):
        """ Score a batch of examples """
        score_list = []
        for i in range(0, len(srcs), batch_size):
            src_list = srcs[i: i + batch_size]
            tgt_list = tgts[i: i + batch_size]
            try:
                with torch.no_grad():
                    encoded_src = self.tokenizer(
                        src_list,
                        max_length=self.max_length,
                        truncation=True,
                        padding=True,
                        return_tensors='pt'
                    )
                    encoded_tgt = self.tokenizer(
                        tgt_list,
                        max_length=self.max_length,
                        truncation=True,
                        padding=True,
                        return_tensors='pt'
                    )
                    src_tokens = encoded_src['input_ids'].to(self.device)
                    src_mask = encoded_src['attention_mask'].to(self.device)

                    tgt_tokens = encoded_tgt['input_ids'].to(self.device)
                    tgt_mask = encoded_tgt['attention_mask']
                    tgt_len = tgt_mask.sum(dim=1).to(self.device)

                    output = self.model(
                        input_ids=src_tokens,
                        attention_mask=src_mask,
                        labels=tgt_tokens
                    )
                    logits = output.logits.view(-1, self.model.config.vocab_size)
                    loss = self.loss_fct(self.lsm(logits), tgt_tokens.view(-1))
                    loss = loss.view(tgt_tokens.shape[0], -1)
                    loss = loss.sum(dim=1) / tgt_len
                    curr_score_list = [-x.item() for x in loss]
                    score_list += curr_score_list

            except RuntimeError:
                traceback.print_exc()
                print(f'source: {src_list}')
                print(f'target: {tgt_list}')
                exit(0)
        return score_list

    def multi_ref_score(self, srcs, tgts: List[List[str]], agg="mean", batch_size=4):
        # Assert we have the same number of references
        ref_nums = [len(x) for x in tgts]
        if len(set(ref_nums)) > 1:
            raise Exception("You have different number of references per test sample.")

        ref_num = len(tgts[0])
        score_matrix = []
        for i in range(ref_num):
            curr_tgts = [x[i] for x in tgts]
            scores = self.score(srcs, curr_tgts, batch_size)
            score_matrix.append(scores)
        if agg == "mean":
            score_list = np.mean(score_matrix, axis=0)
        elif agg == "max":
            score_list = np.max(score_matrix, axis=0)
        else:
            raise NotImplementedError
        return list(score_list)

    def test(self, batch_size=3):
        """ Test """
        src_list = [
            'This is a very good idea. Although simple, but very insightful.',
            'Can I take a look?',
            'Do not trust him, he is a liar.'
        ]

        tgt_list = [
            "That's stupid.",
            "What's the problem?",
            'He is trustworthy.'
        ]

        print(self.score(src_list, tgt_list, batch_size))




import logging
import multiprocessing as mp
from collections.abc import Iterable
from functools import partial
from typing import List

from numpy import random
from omegaconf import DictConfig

log = logging.getLogger(__name__)


CUDA_DEVICES = mp.Queue()
WORKER_CUDA_DEVICES = None


def initialize_worker(num_cuda_devices=1):
    global CUDA_DEVICES
    global WORKER_CUDA_DEVICES
    WORKER_CUDA_DEVICES = []
    for i in range(num_cuda_devices):
        WORKER_CUDA_DEVICES.append(str(CUDA_DEVICES.get()))
    log.info(f"Worker cuda devices: {','.join(WORKER_CUDA_DEVICES)}")


def repeat_tasks(tasks: List[DictConfig]) -> List[DictConfig]:
    rep_tasks: List[DictConfig] = []
    for task in tasks:
        if "seeds" in task:
            seeds = task.seeds
        elif "n_repeats" in tasks and task.n_repeats:
            seeds = random.randint(0, 2147483648, task.n_repeats).tolist()
        else:
            seeds = [random.randint(2147483648)]
        log.info(f"N repeats: {len(seeds)}")
        for i, seed in enumerate(seeds):
            new_task = DictConfig(task)
            new_task.name = f"{new_task.name}_rep{i}"
            new_task.repeat = f"rep{i}"
            new_task.args += f" seed={seed} +fixed_seed=True"
            if "ens_seeds" in task:
                ens_seeds_str = "[" + ",".join(str(e) for e in task.ens_seeds[i]) + "]"
                new_task.args += f" al.ens_seeds={ens_seeds_str}"
            rep_tasks.append(new_task)

    return rep_tasks


def run_tasks(config, f_task):
    global CUDA_QUEUE

    n_cudas_per_run = getattr(config, "cuda_devices_per_run", 1)

    if not isinstance(config.cuda_devices, Iterable):
        cuda_devices = [config.cuda_devices]
    else:
        cuda_devices = config.cuda_devices

    log.info(f"Cuda devices: {cuda_devices}")

    for cuda_device in cuda_devices:
        CUDA_DEVICES.put(cuda_device)

    log.info("All tasks: {}".format(str([t.name for t in config.tasks])))
    if "task_names" in config and config.task_names:
        task_names = config.task_names

        task_index = {t.name: t for t in config.tasks}

        tasks = []
        for task_name in task_names:
            task_name = task_name.split("@")
            if task_name[0] not in task_index:
                raise ValueError(f"Task name: {task_name[0]} is not in config file.")

            task = task_index[task_name[0]]
            if len(task_name) == 2:
                task.n_repeats = int(task_name[1])

            tasks.append(task)
    else:
        tasks = config.tasks

    log.info("Running tasks: {}".format(str([t.name for t in tasks])))
    log.info(f"Cuda devices per run: {n_cudas_per_run}")

    tasks = repeat_tasks(tasks)

    pool = mp.Pool(
        processes=len(cuda_devices) // n_cudas_per_run,
        initializer=partial(initialize_worker, num_cuda_devices=n_cudas_per_run),
    )
    try:
        pool.map(f_task, tasks)

        pool.close()
        pool.join()
    except KeyboardInterrupt:
        pool.terminate()
        pool.join()



import logging

import numpy as np

from .annotation_converter import AnnotationConverter

logger = logging.getLogger()


class AnnotationConverterBio(AnnotationConverter):
    """ Supports span annotator for sequence tagging. """

    def __init__(self, offsets):
        """
        offsets: offests of starting positions of tokens
        """

        self._offsets = np.asarray(offsets)

    def __call__(self, indexes, answers):
        selected_offsets = self._offsets[indexes]

        res = []
        for i in range(selected_offsets.size):
            res.append(
                self.to_bio(answers[i], selected_offsets[i], len(selected_offsets[i]))
            )

        return res

    def find_in_between(self, offsets, start, end):
        res = []
        for i, offset in enumerate(offsets):
            if start <= offset and offset <= end:
                res.append(i)

        return res

    def to_bio(self, y, offsets, n_tokens):
        good_ys = ["O"] * n_tokens

        if not y or y[0] == "None":
            return good_ys

        for w_y in y:
            positions = self.find_in_between(offsets, w_y["start"], w_y["end"])
            good_ys[positions[0]] = "B-" + w_y["tag"]

            for pos in positions[1:]:
                good_ys[pos] = "I-" + w_y["tag"]

        return good_ys



from torch import nn
from torch.nn.utils import spectral_norm, remove_spectral_norm
from torch.nn.utils.spectral_norm import SpectralNorm
from transformers import (
    BertForSequenceClassification,
    DistilBertForSequenceClassification,
    RobertaForSequenceClassification,
    ElectraForSequenceClassification,
    XLNetForSequenceClassification,
    BertForTokenClassification,
    DistilBertForTokenClassification,
    RobertaForTokenClassification,
    ElectraForTokenClassification,
    XLNetForTokenClassification,
)

from ..models import TextClassificationCNN


class Encoder(nn.Module):
    def __init__(self, model, model_name, use_pooling=True):
        super().__init__()
        self.name = model_name
        self.model = next(model.children())
        if model_name == "bert":
            self.model = nn.Sequential(*list(self.model.children())[:-1])
        self.use_pooling = use_pooling

    def forward(self, *args, **kwargs):
        out = self.model(*args, **kwargs)
        if self.use_pooling:
            return out.last_hidden_state[:, 0, :]
        return out.last_hidden_state


class ModelForFeaturesExtraction(nn.Sequential):
    def forward(self, *args, **kwargs):
        for module in self._modules.values():
            break
        inputs = module(*args, **kwargs)
        for i, module in enumerate(self._modules.values()):
            if i:
                inputs = module(inputs)
        return inputs


def get_model_without_cls_layer(
    model, use_activation=False, use_spectralnorm=False, use_pooling=True
):
    model_name = get_name(model)
    model.eval()

    if model_name in ["text_cnn", "custom"]:
        import copy

        model_without_cls_layer = copy.deepcopy(model)
        model_without_cls_layer.return_embeddings = True
    else:
        model_without_cls_layer = ModelForFeaturesExtraction(
            Encoder(model, model_name, use_pooling)
        )
    _add_modules(
        model_without_cls_layer=model_without_cls_layer,
        original_model=model,
        use_activation=use_activation,
        use_spectralnorm=use_spectralnorm,
    )

    return model_without_cls_layer


def _add_modules(
    model_without_cls_layer, original_model, use_activation=False, use_spectralnorm=True
):
    pre_classifier, dropout, activation = _get_pre_classifier_dropout_activation(
        original_model
    )
    if pre_classifier is not None:
        model_without_cls_layer.add_module("dropout", dropout)
        pre_classifier = _modify_layer_spectral_norm(
            pre_classifier, enable=use_spectralnorm
        )
        model_without_cls_layer.add_module("pre_classifier", pre_classifier)
        if use_activation:
            model_without_cls_layer.add_module("activation", activation)


def _modify_layer_spectral_norm(layer, enable=True):
    """
    Function to avoid overriding the layer data
    """
    # Check if spectral norm is already integrated
    name = "weight"
    sn_already_integrated = False
    for k, hook in layer._forward_pre_hooks.items():
        if isinstance(hook, SpectralNorm) and hook.name == name:
            sn_already_integrated = True
            break

    if (sn_already_integrated and not enable) or (not sn_already_integrated and enable):
        layer_copy = nn.Linear(layer.in_features, layer.out_features)
        layer_copy.weight.data = layer.weight.data.clone()
        layer_copy.bias.data = layer.bias.data.clone()
        if enable:
            return spectral_norm(layer_copy)
        return layer_copy
    return layer


def _get_dim(model_without_cls_layer):
    try:
        return model_without_cls_layer.embed_dim
    except:
        return list(model_without_cls_layer.state_dict().items())[-1][1].shape[0]


def _get_pre_classifier_dropout_activation(model):
    if isinstance(model, (BertForSequenceClassification, BertForTokenClassification)):
        return model.bert.pooler.dense, nn.Identity(), nn.Tanh()
    elif isinstance(model, DistilBertForSequenceClassification):
        return model.pre_classifier, nn.Identity(), nn.ReLU()
    # Since `DistilBertForTokenClassification` does not have a pre-classifier
    elif isinstance(model, DistilBertForTokenClassification):
        return None, nn.Identity(), nn.ReLU()
    elif isinstance(
        model, (RobertaForSequenceClassification, RobertaForTokenClassification)
    ):
        return model.classifier.dense, model.classifier.dropout, nn.Tanh()
    elif isinstance(
        model, (ElectraForSequenceClassification, ElectraForTokenClassification)
    ):
        return model.classifier.dense, model.classifier.dropout, nn.GELU()
    elif isinstance(
        model, (XLNetForSequenceClassification, XLNetForTokenClassification)
    ):
        return (
            model.sequence_summary.summary,
            model.sequence_summary.first_dropout,
            nn.Tanh(),
        )
    elif isinstance(model, TextClassificationCNN):
        return None, model.dropout, nn.Identity()
    else:
        try:
            return model.get_pre_classifier_dropout_activation()
        except:
            raise NotImplementedError


def get_name(model):
    if isinstance(model, (BertForSequenceClassification, BertForTokenClassification)):
        return "bert"
    elif isinstance(
        model, (DistilBertForSequenceClassification, DistilBertForTokenClassification)
    ):
        return "distilbert"
    elif isinstance(
        model, (RobertaForSequenceClassification, RobertaForTokenClassification)
    ):
        return "roberta"
    elif isinstance(
        model, (ElectraForSequenceClassification, ElectraForTokenClassification)
    ):
        return "google/electra"
    elif isinstance(
        model, (XLNetForSequenceClassification, XLNetForTokenClassification)
    ):
        return "xlnet"
    elif isinstance(model, TextClassificationCNN):
        return "text_cnn"
    else:
        return "custom"


def _try_remove_spectral_norm(layer):
    try:
        # remove SN, if it used on train
        remove_spectral_norm(layer)
    except:
        pass



import copy
import datetime
import inspect
import os
import shutil
import sys
import time
from pathlib import Path
from typing import Union

import torch
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.optim.sgd import SGD
from torch.utils.data.dataset import ConcatDataset

try:
    from apex import amp
except ImportError:
    amp = None

import flair
import flair.nn
from flair.data import MultiCorpus, Corpus
from flair.datasets import DataLoader
from flair.optim import ExpAnnealLR
from flair.training_utils import (
    init_output_file,
    log_line,
    add_file_handler,
    Result,
    store_embeddings,
)
from flair.models import SequenceTagger
import random

import logging

log = logging.getLogger("flair")


class FlairModelTrainer:
    def __init__(
        self,
        model: flair.nn.Model,
        corpus: Corpus,
        optimizer: torch.optim.Optimizer = SGD,
        epoch: int = 0,
        use_tensorboard: bool = False,
        device: str = "cpu",
        golden_label_type: str = "label_name",
    ):
        """
        Initialize a model trainer
        :param model: The model that you want to train. The model should inherit from flair.nn.Model
        :param corpus: The dataset used to train the model, should be of type Corpus
        :param optimizer: The optimizer to use (typically SGD or Adam)
        :param epoch: The starting epoch (normally 0 but could be higher if you continue training model)
        :param use_tensorboard: If True, writes out tensorboard information
        """
        self.model: flair.nn.Model = model
        self.corpus: Corpus = corpus
        self.optimizer: torch.optim.Optimizer = optimizer
        self.epoch: int = epoch
        self.use_tensorboard: bool = use_tensorboard
        flair.device = torch.device(device)
        self.golden_label_type = golden_label_type

    def train(
        self,
        base_path: Union[Path, str],
        learning_rate: float = 0.1,
        mini_batch_size: int = 32,
        mini_batch_chunk_size: int = None,
        max_epochs: int = 100,
        anneal_factor: float = 0.5,
        patience: int = 3,
        min_learning_rate: float = 0.0001,
        train_with_dev: bool = False,
        monitor_train: bool = False,
        monitor_test: bool = False,
        embeddings_storage_mode: str = "cpu",
        checkpoint: bool = False,
        save_final_model: bool = True,
        anneal_with_restarts: bool = True,
        anneal_with_prestarts: bool = False,
        batch_growth_annealing: bool = False,
        shuffle: bool = True,
        param_selection_mode: bool = False,
        num_workers: int = 0,
        sampler=None,
        use_amp: bool = False,
        amp_opt_level: str = "O1",
        eval_on_train_fraction=0.0,
        eval_on_train_shuffle=False,
        **kwargs,
    ) -> dict:
        """
        Trains any class that implements the flair.nn.Model interface.
        :param base_path: Main path to which all output during training is logged and models are saved
        :param learning_rate: Initial learning rate
        :param mini_batch_size: Size of mini-batches during training
        :param mini_batch_chunk_size: If mini-batches are larger than this number, they get broken down into chunks of this size for processing purposes
        :param max_epochs: Maximum number of epochs to train. Terminates training if this number is surpassed.
        :param anneal_factor: The factor by which the learning rate is annealed
        :param patience: Patience is the number of epochs with no improvement the Trainer waits
         until annealing the learning rate
        :param min_learning_rate: If the learning rate falls below this threshold, training terminates
        :param train_with_dev: If True, training is performed using both train+dev data
        :param monitor_train: If True, training data is evaluated at end of each epoch
        :param monitor_test: If True, test data is evaluated at end of each epoch
        :param embeddings_storage_mode: One of 'none' (all embeddings are deleted and freshly recomputed),
        'cpu' (embeddings are stored on CPU) or 'gpu' (embeddings are stored on GPU)
        :param checkpoint: If True, a full checkpoint is saved at end of each epoch
        :param save_final_model: If True, final model is saved
        :param anneal_with_restarts: If True, the last best model is restored when annealing the learning rate
        :param shuffle: If True, data is shuffled during training
        :param param_selection_mode: If True, testing is performed against dev data. Use this mode when doing
        parameter selection.
        :param num_workers: Number of workers in your data loader.
        :param sampler: You can pass a data sampler here for special sampling of data.
        :param eval_on_train_fraction: the fraction of train data to do the evaluation on,
        if 0. the evaluation is not performed on fraction of training data,
        if 'dev' the size is determined from dev set size
        :param eval_on_train_shuffle: if True the train data fraction is determined on the start of training
        and kept fixed during training, otherwise it's sampled at beginning of each epoch
        :param kwargs: Other arguments for the Optimizer
        :return:
        """

        if self.use_tensorboard:
            try:
                from torch.utils.tensorboard import SummaryWriter

                writer = SummaryWriter()
            except:
                log_line(log)
                log.warning(
                    "ATTENTION! PyTorch >= 1.1.0 and pillow are required for TensorBoard support!"
                )
                log_line(log)
                self.use_tensorboard = False
                pass

        if use_amp:
            if sys.version_info < (3, 0):
                raise RuntimeError("Apex currently only supports Python 3. Aborting.")
            if amp is None:
                raise RuntimeError(
                    "Failed to import apex. Please install apex from https://www.github.com/nvidia/apex "
                    "to enable mixed-precision training."
                )

        if mini_batch_chunk_size is None:
            mini_batch_chunk_size = mini_batch_size

        # cast string to Path
        if type(base_path) is str:
            base_path = Path(base_path)
        else:
            base_path = Path("./")

        log_handler = add_file_handler(log, base_path / "training.log")

        log_line(log)
        log.info(f'Model: "{self.model}"')
        log_line(log)
        log.info(f'Corpus: "{self.corpus}"')
        log_line(log)
        log.info("Parameters:")
        log.info(f' - learning_rate: "{learning_rate}"')
        log.info(f' - mini_batch_size: "{mini_batch_size}"')
        log.info(f' - patience: "{patience}"')
        log.info(f' - anneal_factor: "{anneal_factor}"')
        log.info(f' - max_epochs: "{max_epochs}"')
        log.info(f' - shuffle: "{shuffle}"')
        log.info(f' - train_with_dev: "{train_with_dev}"')
        log.info(f' - batch_growth_annealing: "{batch_growth_annealing}"')
        log_line(log)
        log.info(f'Model training base path: "{base_path}"')
        log_line(log)
        log.info(f"Device: {flair.device}")
        log_line(log)
        log.info(f"Embeddings storage mode: {embeddings_storage_mode}")
        if (
            isinstance(self.model, SequenceTagger)
            and self.model.weight_dict
            and self.model.use_crf
        ):
            log_line(log)
            log.warning(
                f"WARNING: Specified class weights will not take effect when using CRF"
            )

        # determine what splits (train, dev, test) to evaluate and log
        log_train = True if monitor_train else False
        log_test = (
            True
            if (not param_selection_mode and self.corpus.test and monitor_test)
            else False
        )
        log_dev = True if not train_with_dev else False
        log_train_part = (
            True
            if (eval_on_train_fraction == "dev" or eval_on_train_fraction > 0.0)
            else False
        )

        if log_train_part:
            train_part_size = (
                len(self.corpus.dev)
                if eval_on_train_fraction == "dev"
                else int(len(self.corpus.train) * eval_on_train_fraction)
            )
            assert train_part_size > 0
            if not eval_on_train_shuffle:
                train_part_indices = list(range(train_part_size))
                train_part = torch.utils.data.dataset.Subset(
                    self.corpus.train, train_part_indices
                )

        # prepare loss logging file and set up header
        loss_txt = init_output_file(base_path, "loss.tsv")

        # weight_extractor = WeightExtractor(base_path)

        # ADDED MOMENTUM AS IN LIPTON CODE
        optimizer: torch.optim.Optimizer = self.optimizer(
            self.model.parameters(), lr=learning_rate, momentum=0.9, **kwargs
        )

        if use_amp:
            self.model, optimizer = amp.initialize(
                self.model, optimizer, opt_level=amp_opt_level
            )

        # minimize training loss if training with dev data, else maximize dev score
        anneal_mode = "min" if train_with_dev else "max"

        scheduler_lambda = lambda epoch: (1 + anneal_factor * epoch) / (
            1 + anneal_factor * (epoch + 1)
        )
        # scheduler = MultiplicativeLR(optimizer, scheduler_lambda)

        scheduler: ReduceLROnPlateau = ReduceLROnPlateau(
            optimizer,
            factor=anneal_factor,
            patience=patience,
            mode=anneal_mode,
            verbose=True,
        )

        train_data = self.corpus.train

        # if training also uses dev data, include in training set
        if train_with_dev:
            train_data = ConcatDataset([self.corpus.train, self.corpus.dev])

        # initialize sampler if provided
        if sampler is not None:
            # init with default values if only class is provided
            if inspect.isclass(sampler):
                sampler = sampler()
            # set dataset to sample from
            sampler.set_dataset(train_data)
            shuffle = False

        dev_score_history = []
        dev_loss_history = []
        train_loss_history = []

        micro_batch_size = mini_batch_chunk_size

        # before training remove checkpoints from previous iterations, if exist
        dirs_list = os.listdir(base_path)
        dirs_list = [path for path in dirs_list if "checkpoint" in path]
        for directory in dirs_list:
            shutil.rmtree(os.path.join(base_path, directory), ignore_errors=True)
        # At any point you can hit Ctrl + C to break out of training early.
        try:
            previous_learning_rate = learning_rate

            for self.epoch in range(self.epoch + 1, max_epochs + 1):
                log_line(log)

                last_epoch_model_state_dict = self.model.state_dict()
                best_state_dict = self.model.state_dict()

                if eval_on_train_shuffle:
                    train_part_indices = list(range(self.corpus.train))
                    random.shuffle(train_part_indices)
                    train_part_indices = train_part_indices[:train_part_size]
                    train_part = torch.utils.data.dataset.Subset(
                        self.corpus.train, train_part_indices
                    )

                # get new learning rate
                for group in optimizer.param_groups:
                    learning_rate = group["lr"]

                if learning_rate != previous_learning_rate and batch_growth_annealing:
                    mini_batch_size *= 2

                # reload last best model if annealing with restarts is enabled
                if (
                    learning_rate
                    != previous_learning_rate
                    # and (base_path / "best-model.pt").exists()
                ):
                    log.info("resetting to best model")
                    if anneal_with_restarts:
                        self.model.load_state_dict(current_best_state_dict)
                    if anneal_with_prestarts:
                        self.model.load_state_dict(
                            self.model.load(
                                base_path / "pre-best-model.pt"
                            ).state_dict()
                        )

                previous_learning_rate = learning_rate

                # stop training if learning rate becomes too small
                if learning_rate < min_learning_rate:
                    log_line(log)
                    log.info("learning rate too small - quitting training!")
                    log_line(log)
                    break

                batch_loader = DataLoader(
                    train_data,
                    batch_size=mini_batch_size,
                    shuffle=shuffle,
                    num_workers=num_workers,
                    sampler=sampler,
                )

                self.model.train()

                train_loss: float = 0

                seen_batches = 0
                total_number_of_batches = len(batch_loader)

                modulo = max(1, int(total_number_of_batches / 10))

                # process mini-batches
                batch_time = 0
                for batch_no, batch in enumerate(batch_loader):
                    start_time = time.time()

                    # zero the gradients on the model and optimizer
                    self.model.zero_grad()
                    optimizer.zero_grad()

                    # if necessary, make batch_steps
                    batch_steps = [batch]
                    if len(batch) > micro_batch_size:
                        batch_steps = [
                            batch[x : x + micro_batch_size]
                            for x in range(0, len(batch), micro_batch_size)
                        ]

                    # forward and backward for batch
                    for batch_step in batch_steps:

                        # forward pass
                        # CHANGED!!!
                        # loss = self.model.forward_loss(batch_step)
                        loss = self.model.forward_loss(batch_step)[0]

                        # Backward
                        if use_amp:
                            with amp.scale_loss(loss, optimizer) as scaled_loss:
                                scaled_loss.backward()
                        else:
                            loss.backward()

                    # do the optimizer step
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), 5.0)
                    optimizer.step()

                    seen_batches += 1
                    train_loss += loss.item()

                    # depending on memory mode, embeddings are moved to CPU, GPU or deleted
                    store_embeddings(batch, embeddings_storage_mode)

                    batch_time += time.time() - start_time
                    if seen_batches % modulo == 0:
                        log.info(
                            f"epoch {self.epoch} - iter {seen_batches}/{total_number_of_batches} - loss "
                            f"{train_loss / seen_batches / mini_batch_size:.8f} - samples/sec: {mini_batch_size * modulo / batch_time:.2f}"
                        )
                        batch_time = 0
                        iteration = self.epoch * total_number_of_batches + batch_no
                #                         if not param_selection_mode:
                #                             weight_extractor.extract_weights(
                #                                 self.model.state_dict(), iteration
                #                             )

                # slightly change loss calculation - divide on batch size
                train_loss /= seen_batches * mini_batch_size

                self.model.eval()

                log_line(log)
                log.info(
                    f"EPOCH {self.epoch} done: loss {train_loss:.4f} - lr {learning_rate:.7f}"
                )

                if self.use_tensorboard:
                    writer.add_scalar("train_loss", train_loss, self.epoch)

                # anneal against train loss if training with dev, otherwise anneal against dev score
                current_score = train_loss

                # evaluate on train / dev / test split depending on training settings
                result_line: str = ""

                if log_train:
                    train_eval_result, train_loss = self.model.evaluate(
                        self.corpus.train,
                        self.golden_label_type,
                        embedding_storage_mode=embeddings_storage_mode,
                        mini_batch_size=mini_batch_chunk_size,
                        num_workers=num_workers,
                    )
                    result_line += f"\t{train_eval_result.log_line}"

                    # depending on memory mode, embeddings are moved to CPU, GPU or deleted
                    store_embeddings(self.corpus.train, embeddings_storage_mode)

                if log_train_part:
                    train_part_eval_result, train_part_loss = self.model.evaluate(
                        train_part,
                        self.golden_label_type,
                        embedding_storage_mode=embeddings_storage_mode,
                        mini_batch_size=mini_batch_chunk_size,
                        num_workers=num_workers,
                    )
                    result_line += (
                        f"\t{train_part_loss}\t{train_part_eval_result.log_line}"
                    )
                    log.info(
                        f"TRAIN_SPLIT : loss {train_part_loss} - score {round(train_part_eval_result.main_score, 4)}"
                    )

                if log_dev:
                    dev_eval_result = self.model.evaluate(
                        self.corpus.dev,
                        self.golden_label_type,
                        embedding_storage_mode=embeddings_storage_mode,
                        mini_batch_size=mini_batch_chunk_size,
                        num_workers=num_workers,
                    )
                    dev_loss = dev_eval_result.loss
                    result_line += f"\t{dev_loss}\t{dev_eval_result.log_line}"
                    log.info(
                        f"DEV : loss {dev_loss} - score {round(dev_eval_result.main_score, 4)}"
                    )
                    # calculate scores using dev data if available
                    # append dev score to score history
                    dev_score_history.append(dev_eval_result.main_score)
                    dev_loss_history.append(dev_loss)

                    current_score = dev_eval_result.main_score

                    # depending on memory mode, embeddings are moved to CPU, GPU or deleted
                    store_embeddings(self.corpus.dev, embeddings_storage_mode)

                    if self.use_tensorboard:
                        writer.add_scalar("dev_loss", dev_loss, self.epoch)
                        writer.add_scalar(
                            "dev_score", dev_eval_result.main_score, self.epoch
                        )

                if log_test:
                    test_eval_result, test_loss = self.model.evaluate(
                        self.corpus.test,
                        self.golden_label_type,
                        base_path / "test.tsv",
                        embedding_storage_mode=embeddings_storage_mode,
                        mini_batch_size=mini_batch_chunk_size,
                        num_workers=num_workers,
                    )
                    result_line += f"\t{test_loss}\t{test_eval_result.log_line}"
                    log.info(
                        f"TEST : loss {test_loss} - score {round(test_eval_result.main_score, 4)}"
                    )

                    # depending on memory mode, embeddings are moved to CPU, GPU or deleted
                    store_embeddings(self.corpus.test, embeddings_storage_mode)

                    if self.use_tensorboard:
                        writer.add_scalar("test_loss", test_loss, self.epoch)
                        writer.add_scalar(
                            "test_score", test_eval_result.main_score, self.epoch
                        )

                # determine learning rate annealing through scheduler
                scheduler.step(current_score)
                # scheduler.step()

                train_loss_history.append(train_loss)

                # determine bad epoch number
                try:
                    bad_epochs = scheduler.num_bad_epochs
                except:
                    bad_epochs = 0
                for group in optimizer.param_groups:
                    new_learning_rate = group["lr"]
                if new_learning_rate != previous_learning_rate:
                    bad_epochs = patience + 1

                # log bad epochs
                log.info(f"BAD EPOCHS (no improvement): {bad_epochs}")

                # output log file
                with open(loss_txt, "a") as f:

                    # make headers on first epoch
                    if self.epoch == 1:
                        f.write(
                            f"EPOCH\tTIMESTAMP\tBAD_EPOCHS\tLEARNING_RATE\tTRAIN_LOSS"
                        )

                        if log_train:
                            f.write(
                                "\tTRAIN_"
                                + "\tTRAIN_".join(
                                    train_eval_result.log_header.split("\t")
                                )
                            )
                        if log_train_part:
                            f.write(
                                "\tTRAIN_PART_LOSS\tTRAIN_PART_"
                                + "\tTRAIN_PART_".join(
                                    train_part_eval_result.log_header.split("\t")
                                )
                            )
                        if log_dev:
                            f.write(
                                "\tDEV_LOSS\tDEV_"
                                + "\tDEV_".join(dev_eval_result.log_header.split("\t"))
                            )
                        if log_test:
                            f.write(
                                "\tTEST_LOSS\tTEST_"
                                + "\tTEST_".join(
                                    test_eval_result.log_header.split("\t")
                                )
                            )

                    f.write(
                        f"\n{self.epoch}\t{datetime.datetime.now():%H:%M:%S}\t{bad_epochs}\t{learning_rate:.4f}\t{train_loss}"
                    )
                    f.write(result_line)

                # if checkpoint is enabled, save model at each epoch
                if checkpoint and not param_selection_mode:
                    save_path = os.path.join(base_path, f"checkpoint-{self.epoch}")
                    os.makedirs(save_path, exist_ok=True)
                    save_path = os.path.join(save_path, "pytorch_model.pt")
                    self.model.save(save_path)
                    # self.save_checkpoint(base_path / "checkpoint.pt")
                    # remove older checkpoints
                    try:
                        older_path = os.path.join(
                            base_path, f"checkpoint-{self.epoch - patience}"
                        )
                        shutil.rmtree(older_path, ignore_errors=True)
                    except:
                        pass

                # if we use dev data, remember best model based on dev evaluation score
                if (
                    (
                        not train_with_dev
                        or anneal_with_restarts
                        or anneal_with_prestarts
                    )
                    and not param_selection_mode
                    # REMOVE CAUSE NEW SCHEDULER
                    and current_score == scheduler.best
                ):
                    print("saving best model")
                    # self.model.save(base_path / "best-model.pt")
                    current_best_state_dict = copy.deepcopy(self.model.state_dict())

                    if anneal_with_prestarts:
                        self.model.load_state_dict(last_epoch_model_state_dict)
                        self.model.save(base_path / "pre-best-model.pt")
                    # self.model.load_state_dict(current_state_dict)

            # if we do not use dev data for model selection, save final model
            if save_final_model and not param_selection_mode:
                save_path = base_path
                os.makedirs(save_path, exist_ok=True)
                save_path = os.path.join(save_path, "final-model.pt")
                self.model.save(save_path)

        except KeyboardInterrupt:
            log_line(log)
            log.info("Exiting from training early.")

            if self.use_tensorboard:
                writer.close()

            if not param_selection_mode:
                log.info("Saving model ...")
                save_path = base_path
                os.makedirs(save_path, exist_ok=True)
                save_path = os.path.join(save_path, "final-model.pt")
                self.model.save(save_path)
                log.info("Done.")

        # test best model if test data is present
        if self.corpus.test:
            final_score = self.final_test(base_path, mini_batch_chunk_size, num_workers)
        else:
            final_score = 0
            log.info("Test data not provided setting final score to 0")

        log.removeHandler(log_handler)

        if self.use_tensorboard:
            writer.close()

        return {
            "test_score": final_score,
            "dev_score_history": dev_score_history,
            "train_loss_history": train_loss_history,
            "dev_loss_history": dev_loss_history,
        }

    def save_checkpoint(self, model_file: Union[str, Path]):
        corpus = self.corpus
        self.corpus = None
        torch.save(self, str(model_file), pickle_protocol=4)
        self.corpus = corpus

    @classmethod
    def load_checkpoint(cls, checkpoint: Union[Path, str], corpus: Corpus):
        model: FlairModelTrainer = torch.load(checkpoint, map_location=flair.device)
        model.corpus = corpus
        return model

    def final_test(
        self, base_path: Path, eval_mini_batch_size: int, num_workers: int = 8
    ):

        log_line(log)
        log.info("Testing using best model ...")

        self.model.eval()

        if (base_path / "best-model.pt").exists():
            self.model = self.model.load(base_path / "best-model.pt")

        test_results, test_loss = self.model.evaluate(
            self.corpus.test,
            self.golden_label_type,
            out_path=base_path / "test.tsv",
            embedding_storage_mode="none",
            mini_batch_size=eval_mini_batch_size,
            num_workers=num_workers,
        )

        test_results: Result = test_results
        log.info(test_results.log_line)
        log.info(test_results.detailed_results)
        log_line(log)

        # if we are training over multiple datasets, do evaluation for each
        if type(self.corpus) is MultiCorpus:
            for subcorpus in self.corpus.corpora:
                log_line(log)
                self.model.evaluate(
                    subcorpus.test,
                    self.golden_label_type,
                    out_path=base_path / f"{subcorpus.name}-test.tsv",
                    embedding_storage_mode="none",
                    mini_batch_size=eval_mini_batch_size,
                    num_workers=num_workers,
                )

        # get and return the final test score of best model
        final_score = test_results.main_score

        return final_score

    def find_learning_rate(
        self,
        base_path: Union[Path, str],
        file_name: str = "learning_rate.tsv",
        start_learning_rate: float = 1e-7,
        end_learning_rate: float = 10,
        iterations: int = 100,
        mini_batch_size: int = 32,
        stop_early: bool = True,
        smoothing_factor: float = 0.98,
        **kwargs,
    ) -> Path:
        best_loss = None
        moving_avg_loss = 0

        # cast string to Path
        if type(base_path) is str:
            base_path = Path(base_path)
        learning_rate_tsv = init_output_file(base_path, file_name)

        with open(learning_rate_tsv, "a") as f:
            f.write("ITERATION\tTIMESTAMP\tLEARNING_RATE\tTRAIN_LOSS\n")

        optimizer = self.optimizer(
            self.model.parameters(), lr=start_learning_rate, **kwargs
        )

        train_data = self.corpus.train

        scheduler = ExpAnnealLR(optimizer, end_learning_rate, iterations)

        model_state = self.model.state_dict()
        self.model.train()

        step = 0
        while step < iterations:
            batch_loader = DataLoader(
                train_data, batch_size=mini_batch_size, shuffle=True
            )
            for batch in batch_loader:
                step += 1

                # forward pass
                loss = self.model.forward_loss(batch)

                # update optimizer and scheduler
                optimizer.zero_grad()
                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), 5.0)
                optimizer.step()
                scheduler.step(step)

                print(scheduler.get_lr())
                learning_rate = scheduler.get_lr()[0]

                loss_item = loss.item()
                if step == 1:
                    best_loss = loss_item
                else:
                    if smoothing_factor > 0:
                        moving_avg_loss = (
                            smoothing_factor * moving_avg_loss
                            + (1 - smoothing_factor) * loss_item
                        )
                        loss_item = moving_avg_loss / (
                            1 - smoothing_factor ** (step + 1)
                        )
                    if loss_item < best_loss:
                        best_loss = loss

                if step > iterations:
                    break

                if stop_early and (loss_item > 4 * best_loss or torch.isnan(loss)):
                    log_line(log)
                    log.info("loss diverged - stopping early!")
                    step = iterations
                    break

                with open(str(learning_rate_tsv), "a") as f:
                    f.write(
                        f"{step}\t{datetime.datetime.now():%H:%M:%S}\t{learning_rate}\t{loss_item}\n"
                    )

            self.model.load_state_dict(model_state)
            self.model.to(flair.device)

        log_line(log)
        log.info(f"learning rate finder finished - plot {learning_rate_tsv}")
        log_line(log)

        return Path(learning_rate_tsv)



import math

import numpy as np
import pandas as pd
import torch
from scipy.special import softmax
from sklearn.metrics import accuracy_score
from torch import nn, optim
from torch.nn import functional as f


def _split_into_bins(n_bins, probs, labels):
    bins = []
    true_labels_for_bins = []

    for i in range(n_bins):
        bins.append([])
        true_labels_for_bins.append([])

    for j in range(len(labels)):
        max_p = max(probs[j])
        for i in range(n_bins):
            if (i / n_bins) < max_p <= (i + 1) / n_bins:
                bins[i].append((probs[j]))
                true_labels_for_bins[i].append(labels[j])

    return bins, true_labels_for_bins


def compute_ece(n_bins, probs, labels, len_dataset):
    bins, true_labels_for_bins = _split_into_bins(n_bins, probs, labels)
    bins = list(filter(None, bins))
    true_labels_for_bins = list(filter(None, true_labels_for_bins))
    ece = torch.zeros(1)
    for i in range(len(bins)):
        softmaxes = torch.from_numpy(np.array(bins[i]))
        confidences, predictions = torch.max(softmaxes, dim=1)
        accuracy = accuracy_score(true_labels_for_bins[i], predictions)
        confidence = torch.sum(confidences) / len(bins[i])
        ece += len(bins[i]) * torch.abs(accuracy - confidence) / len_dataset
    return ece


def _split_into_classes(labels, probs):
    class_probs = []
    dict_class_probs = {}
    n_classes = np.shape(probs)[1]
    for i in range(n_classes):
        class_probs.append([])
    for ind, label in enumerate(labels):
        for i in range(n_classes):
            if label == i:
                class_probs[i].append(probs[ind])
    for i in range(n_classes):
        dict_class_probs[i] = class_probs[i]
    return dict_class_probs


def compute_sce(nbins, probs, labels):
    ece_values_for_each_class = []
    dict_class_probs = _split_into_classes(labels, probs)
    for item in dict_class_probs.keys():
        ece_values_for_each_class.append(
            compute_ece(
                nbins,
                dict_class_probs[item],
                np.array([item] * np.shape(dict_class_probs[item])[0]),
                len(labels),
            )
        )
    return sum(ece_values_for_each_class) / len(dict_class_probs.keys())


def _split_into_ranges(R, probs, labels):
    N = len(probs)
    bins = []
    true_labels = []
    for i in range(R):
        bins.append([])
        true_labels.append([])
    for j in range(R):
        for i in range(j * math.floor(N / R), (j + 1) * math.floor(N / R)):
            bins[j].append(probs[i])
            true_labels[j].append(labels[i])
    return np.array(bins), np.array(true_labels)


def compute_ace(R, probs, labels):
    dict_class_probs = _split_into_classes(labels, probs)
    summa = 0
    for item in dict_class_probs.keys():
        class_labels = np.array([item] * np.shape(dict_class_probs[item])[0])
        class_probs = dict_class_probs[item]
        bins, true_labels = _split_into_ranges(R, class_probs, class_labels)
        for binn, bin_labels in zip(bins, true_labels):
            conf_array, predictions = torch.max(torch.from_numpy(binn), dim=1)
            accuracy = accuracy_score(bin_labels, predictions.numpy())
            confidence = torch.sum(conf_array) / len(conf_array)
            substraction = abs(accuracy - confidence)
            summa += substraction
    ACE = summa / (len(dict_class_probs.keys()) * R)
    return ACE


def _choose_data(threshold, probs, labels):
    arr = torch.max(torch.from_numpy(np.array(probs)), dim=1)[0]
    arr.numpy()
    arr_with_indices = list(enumerate(arr))
    arr_with_indices.sort(key=lambda x: x[1])
    thr_array = []
    for pair in arr_with_indices:
        if pair[1] > threshold:
            thr_array.append(pair)
    indices = []
    for pair in thr_array:
        indices.append(pair[0])
    chosen_labels = labels[indices]
    chosen_probs = probs[indices]
    return chosen_labels, chosen_probs


def compute_tace(threshold, probs, labels, R):
    if isinstance(labels, pd.DataFrame) or isinstance(labels, pd.Series):
        labels = labels.to_numpy()
    chosen_labels, chosen_probs = _choose_data(threshold, probs, labels)
    return compute_ace(R, chosen_probs, chosen_labels)


class ModelWithTempScaling(nn.Module):
    """
    A wrapper for a model with temperature scaling
    model: a classification neural network
    n_classes: number of classes in the dataset
    """

    def __init__(self, model):
        super(ModelWithTempScaling, self).__init__()
        self.model = model
        self.temperature = nn.Parameter(torch.ones(1))

    def forward(self, input):
        logits = self.model(input)
        return f.softmax(torch.true_divide(logits, self.temperature), dim=1)

    def scaling(self, logits, labels, lr=0.01, max_iter=50):
        # logits and labels must be from calibration dataset
        nll = nn.CrossEntropyLoss()
        optimizer = optim.LBFGS([self.temperature], lr=lr, max_iter=max_iter)

        def eval():
            loss = nll(torch.true_divide(logits, self.temperature), labels)
            loss.backward()
            return loss

        optimizer.step(eval)
        return self


class ModelWithVectScaling(nn.Module):
    """
    A wrapper for a model with vector scaling
    model:  a classification neural network
    n_classes: number of classes in the dataset
    """

    def __init__(self, model, n_classes):
        super(ModelWithVectScaling, self).__init__()
        self.model = model
        self.W_and_b = nn.Parameter(
            torch.cat((torch.ones(n_classes), torch.zeros(n_classes)), dim=0)
        )

    def forward(self, input):
        logits = self.model(input)
        return f.softmax(self.scaling_logits(logits), dim=1)

    def scaling_logits(self, logits):
        # logits and labels must be from calibration dataset
        W = torch.diag(self.W_and_b[: logits.shape[1]])
        b = self.W_and_b[logits.shape[1] :]
        b = b.unsqueeze(0).expand(logits.shape[0], -1)
        return torch.mm(logits.float(), W) + b

    def scaling(self, logits, labels, lr=0.00001, max_iter=3500):
        nll = nn.CrossEntropyLoss()
        optimizer = optim.LBFGS([self.W_and_b], lr=lr, max_iter=max_iter)

        def eval():
            loss = nll(self.scaling_logits(logits), labels)
            loss.backward()
            return loss

        optimizer.step(eval)
        return self


class ModelWithMatrScaling(nn.Module):
    """
    A wrapper for a model with matrix scaling
    model: a classification neural network
    n_classes: number of classes in the dataset
    """

    def __init__(self, model, n_classes):
        super(ModelWithMatrScaling, self).__init__()
        self.model = model
        self.W = nn.Parameter(torch.diag(torch.ones(n_classes)))
        self.b = nn.Parameter(torch.zeros(n_classes))

    def forward(self, input):
        logits = self.model(input)
        return f.softmax(self.scaling_logits(logits), dim=1)

    def scaling_logits(self, logits):
        self.b.unsqueeze(0).expand(logits.shape[0], -1)
        return torch.mm(logits.float(), self.W) + self.b

    def scaling(self, logits, labels, lr=0.001, max_iter=100):
        # logits and labels must be from calibration dataset
        nll = nn.CrossEntropyLoss()
        optimizer = optim.LBFGS([self.W, self.b], lr=lr, max_iter=max_iter)

        def eval():
            loss = nll(self.scaling_logits(logits), labels)
            loss.backward()
            return loss

        optimizer.step(eval)
        return self


def binary_histogram_binning(num_bins, probs, labels, probs_to_calibrate):
    """
    histogram binning for  binary classification
    :param num_bins: number of bins
    :param probs: probabilities on calibration dataset
    :param labels: labels of calibration dataset
    :param probs_to_calibrate: initial probabilities on test dataset (which need to be calibrated)
    :return: calibrated probabilities on test dataset
    """
    bins = np.linspace(0, 1, num=num_bins)
    indexes_list = np.digitize(probs, bins) - 1
    theta = np.zeros(num_bins)
    for i in range(len(bins)):
        binn = indexes_list == i
        binn_len = np.sum(binn)
        if binn_len != 0:
            theta[i] = np.sum(labels[binn]) / binn_len
        else:
            theta[i] = bins[i]
    return list(map(lambda x: theta[np.digitize(x, bins) - 1], probs_to_calibrate))


def multiclass_histogram_binning(num_bins, logits, labels, logits_to_calibrate):
    """
    histogram binning for multiclass classification
    :param num_bins: number of bins
    :param logits: logits on calibration dataset
    :param labels: labels on calibration dataset
    :param logits_to_calibrate: initial logits on test dataset (which need to be calibrated)
    :return: calibrated probabilities on test dataset
    """
    probs = softmax(logits, axis=1)
    probs_to_calibrate = softmax(logits_to_calibrate, axis=1)
    binning_res = []
    for k in range(np.shape(probs)[1]):
        binning_res.append(
            binary_histogram_binning(
                num_bins, probs[:, k], labels == k, probs_to_calibrate[:, k]
            )
        )
    binning_res = np.vstack(binning_res).T
    cal_confs = binning_res / (np.sum(binning_res, axis=1)[:, None])
    return cal_confs



import logging
from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset
from sklearn.cluster import KMeans
from sklearn.preprocessing import normalize

from .al_strategy_utils import take_idx
from ..utils.get_embeddings import get_embeddings
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def embeddings_km(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    data_is_tokenized=False,
    batch_size=100,
    data_config=None,
    **embeddings_km_kwargs,
):
    """
    Employs the surprisal embedding of w, which is obtained from the likelihoods of masked tokens from
    pre-trained language models. Doesn't require fine-tuning. https://aclanthology.org/2020.emnlp-main.637/
    """
    data_config = data_config if data_config is not None else model.data_config
    kwargs = dict(
        # General
        model=model.model,
        batch_size=batch_size,
        to_numpy=True,
        data_is_tokenized=data_is_tokenized,
        tokenizer=model.tokenizer,
        task=model.task,
        text_name=data_config["text_name"],
        label_name=data_config["label_name"],
    )
    embeddings = get_embeddings(dataloader_or_data=X_pool, **kwargs)
    # embeddings = torch.nn.functional.normalize(embeddings)
    embeddings = normalize(embeddings)

    km = KMeans(n_clusters=n_instances)
    km.fit(embeddings)

    query_idx = get_nearest_to_centers(km.cluster_centers_, embeddings)

    query = X_pool.select(query_idx)

    # Uncertainty estimates are not defined for BERT-KM
    uncertainty_estimates = np.zeros(len(X_pool))

    return query_idx, query, uncertainty_estimates


def get_nearest_to_centers(centers, vectors):
    sim = np.matmul(centers, vectors.T)
    return sim.argmax(axis=1)



from typing import Union
from datasets import Dataset
import numpy as np

from .al_strategy_utils import filter_by_metric, filter_by_uncertainty
from ..utils.transformers_dataset import TransformersDataset


def nsp(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **kwargs,
):
    # Filtering part begin
    filtering_mode = kwargs.get("filtering_mode", None)
    uncertainty_threshold = kwargs.get("uncertainty_threshold", None)
    uncertainty_mode = kwargs.get(
        "uncertainty_mode", "absolute"
    )  # "relative" or "absolute"
    # Filtering part end
    generate_output = model.generate(X_pool, to_numpy=True)
    scores = generate_output["sequences_scores"]
    sequences_ids = generate_output["sequences"]
    # The larger the score, the more confident the model is
    uncertainty_estimates = -scores
    # Filtering part begin
    if filtering_mode == "uncertainty":
        query_idx, uncertainty_estimates = filter_by_uncertainty(
            uncertainty_estimates=uncertainty_estimates,
            uncertainty_threshold=uncertainty_threshold,
            uncertainty_mode=uncertainty_mode,
            n_instances=n_instances,
        )
    elif filtering_mode in ["rouge1", "rouge2", "rougeL", "sacrebleu"]:
        query_idx, uncertainty_estimates = filter_by_metric(
            uncertainty_estimates=uncertainty_estimates,
            uncertainty_threshold=uncertainty_threshold,
            uncertainty_mode=uncertainty_mode,
            n_instances=n_instances,
            texts=X_pool[model.data_config["text_name"]],
            generated_sequences_ids=sequences_ids,
            tokenizer=model.tokenizer,
            metric_cache_dir=model.cache_dir / "metrics",
            metric_name=filtering_mode,
            agg=kwargs.get("filtering_aggregation", "precision"),
        )
    else:
        argsort = np.argsort(-uncertainty_estimates)
        query_idx = argsort[:n_instances]
    # Filtering part end
    query = X_pool.select(query_idx)

    return query_idx, query, uncertainty_estimates



import logging
import multiprocessing.dummy as mp
from multiprocessing import Array

import pandas as pd

from .active_learner import ActiveLearner

logger = logging.getLogger()


class ActiveLearnerAsync(ActiveLearner):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_busy = False
        self.instance_priority = Array("i", list(range(len(self._y_unlabeled_dataset))))
        self._remove_labeled_from_priorities()

    def is_busy(self):
        return self._is_busy

    def choose_samples_for_annotation(self):
        if self._iteration_num < self._rnd_start_steps:
            return self.choose_random_sample_for_annotation(self.n_instances)

        message = {"msg_type": "choose_samples_to_annotate"}
        self.input_queue.put(message)
        return self.instance_priority[: self.n_instances]

    def make_iteration(self, indexes, answers):
        self.add_answers(indexes, answers)
        message = {"msg_type": "make_iteration", "indexes": indexes, "answers": answers}
        self.input_queue.put(message)
        self._remove_labeled_from_priorities()

    def _remove_labeled_from_priorities(self):
        y = pd.Series(
            self._y_unlabeled_dataset, index=list(range(len(self._y_unlabeled_dataset)))
        )
        labeled_indexes = y[y.notnull()].index
        self.instance_priority = [
            e for e in self.instance_priority if e not in labeled_indexes
        ]

    def start(self):
        def process_fn(input_queue, output_queue):
            logger.info("Starting backend loop")

            self._is_busy = True
            self._active_learn_alg.start(
                self._X_labeled_dataset,
                self._y_labeled_dataset,
                self._X_unlabeled_dataset,
                self._y_unlabeled_dataset,
            )
            self._is_busy = False

            self._iteration_num += 1

            while True:
                try:
                    message = input_queue.get(block=True)

                    self._is_busy = True

                    logger.info(f"Received scoring request with {message['msg_type']}")

                    msg_type = message["msg_type"]
                    if msg_type == "choose_samples_to_annotate":
                        indexes = self._active_learn_alg.choose_samples_for_annotation(
                            len(self.instance_priority)
                        )
                        self.instance_priority = indexes
                        response = {"indexes": indexes}

                    elif msg_type == "make_iteration":
                        indexes = message["indexes"]
                        answers = message["answers"]
                        self._active_learn_alg.make_iteration(indexes, answers)
                        response = {"res": 0}
                        self._iteration_num += 1

                        logger.info("Done make iteration. Results are put in queue")

                    else:
                        logger.info(f"Unmatched message type {msg_type}")

                    output_queue.put(response)
                    self._is_busy = False

                except Exception as e:
                    logger.info(f"Exception occurred = {e}")

            logger.info("Process end-of-live")
            return

        self.input_queue = mp.Queue()
        self.output_queue = mp.Queue()

        self.process = mp.Process(
            name="background_process",
            target=process_fn,
            args=(self.input_queue, self.output_queue,),
        )
        self.process.start()



import logging
import os
from pathlib import Path

import hydra
from hydra.core.hydra_config import HydraConfig
from omegaconf import OmegaConf

os.environ["WANDB_DISABLED"] = "true"

log = logging.getLogger()


def main_decorator(func):
    def run_script(config):
        # Set the working directory
        if HydraConfig.initialized():
            auto_generated_dir = os.getcwd()
            os.chdir(hydra.utils.get_original_cwd())
        else:
            auto_generated_dir = config.output_dir
        log.info(f"Work dir: {auto_generated_dir}")
        # Save config into yaml format
        with open(Path(auto_generated_dir) / "config.yaml", "w") as f:
            f.write(OmegaConf.to_yaml(config))
        # Enable offline mode for HuggingFace libraries if necessary
        if config.offline_mode:
            os.environ["HF_DATASETS_OFFLINE"] = "1"
            os.environ["TRANSFORMERS_OFFLINE"] = "1"

        os.environ["PYTHONHASHSEED"] = str(config.seed)
        from transformers import set_seed

        set_seed(config.seed)

        func(config, work_dir=Path(auto_generated_dir))

    return run_script



import os.path
import sys
from pathlib import Path
from copy import deepcopy
from functools import partial
from importlib import import_module

from ..active_learner import ActiveLearner
from ..pool_subsampling_strategies import (
    ups_subsampling,
    random_subsampling,
    naive_subsampling,
)
from ..query_strategies.actune import actune
from ..query_strategies.al_strategy import (
    mahalanobis_triplet_sampling,
    triplet_sampling,
    mahalanobis_filtering_sampling,
    ddu_sampling,
    logits_lc_sampling,
    margin_sampling,
    oracle_sampling,
    hybrid_sampling,
    ssal_sampling,
    ddu_sampling_cv,
)
from ..query_strategies.alps import alps
from ..query_strategies.badge import badge
from ..query_strategies.bait import bait
from ..query_strategies.bald import bald
from ..query_strategies.batchbald import batchbald
from ..query_strategies.breaking_ties import breaking_ties
from ..query_strategies.cal import cal
from ..query_strategies.cluster_margin import cluster_margin
from ..query_strategies.coreset import coreset
from ..query_strategies.egl import egl
from ..query_strategies.embeddings_km import embeddings_km
from ..query_strategies.entropy import entropy
from ..query_strategies.lc import lc
from ..query_strategies.mahalanobis_sampling import mahalanobis_sampling
from ..query_strategies.mnlp import mnlp
from ..query_strategies.random_sampling import random_sampling
from ..query_strategies.strategy_wrappers.small_text_sampling import small_text_sampling
from ..query_strategies.strategy_wrappers.modal_sampling import modal_sampling

QUERY_STRATEGIES = {
    # Classification strategies
    "random": partial(random_sampling, select_by_number_of_tokens=False),
    "entropy": entropy,
    "lc": lc,
    "ssal": ssal_sampling,
    "logits_lc": logits_lc_sampling,
    "br_ties": breaking_ties,
    "mahalanobis": mahalanobis_sampling,
    "mahalanobis_triplet": mahalanobis_triplet_sampling,
    "mahalanobis_filtering": mahalanobis_filtering_sampling,
    "triplet": triplet_sampling,
    "ddu": ddu_sampling,
    "margin": margin_sampling,
    "cal": cal,
    "oracle": oracle_sampling,
    "cluster_margin": cluster_margin,
    "hybrid": hybrid_sampling,
    "alps": alps,
    "coreset": coreset,
    "bait": bait,
    "badge": badge,
    "emb_km": embeddings_km,
    "egl": egl,
    "actune_lc": partial(actune, query_strategy=lc),
    "actune_ent": partial(actune, query_strategy=entropy),
    "actune_cal": partial(actune, query_strategy=cal),
    "actune_mahalanobis": partial(actune, query_strategy=mahalanobis_sampling),
    "small-text_lc": partial(small_text_sampling, small_text_strategy="lc"),
    "small-text_ent": partial(small_text_sampling, small_text_strategy="ent"),
    "modal_lc": partial(modal_sampling, modal_strategy="lc"),
    "modal_ent": partial(modal_sampling, modal_strategy="ent"),
    # NER strategies
    "mnlp_tokens": partial(mnlp, select_by_number_of_tokens=True),
    "mnlp_samples": partial(mnlp, select_by_number_of_tokens=False),
    "random_tokens": partial(random_sampling, select_by_number_of_tokens=True),
    "random_samples": partial(random_sampling, select_by_number_of_tokens=False),
    # BALD
    "bald": partial(bald, select_by_number_of_tokens=False, only_head_dropout=False),
    "bald_head": partial(
        bald, select_by_number_of_tokens=False, only_head_dropout=True
    ),
    "bald_tokens": partial(
        bald, select_by_number_of_tokens=True, only_head_dropout=False
    ),
    "bald_samples": partial(
        bald, select_by_number_of_tokens=False, only_head_dropout=False
    ),
    "bald_tokens_head": partial(
        bald, select_by_number_of_tokens=True, only_head_dropout=True
    ),
    "bald_samples_head": partial(
        bald, select_by_number_of_tokens=False, only_head_dropout=True
    ),
    # BatchBald
    "batchbald": partial(
        batchbald, select_by_number_of_tokens=False, only_head_dropout=False
    ),
    "batchbald_head": partial(
        batchbald, select_by_number_of_tokens=False, only_head_dropout=True
    ),
    "batchbald_tokens": partial(
        batchbald, select_by_number_of_tokens=True, only_head_dropout=False
    ),
    "batchbald_samples": partial(
        batchbald, select_by_number_of_tokens=False, only_head_dropout=False
    ),
    "batchbald_tokens_head": partial(
        batchbald, select_by_number_of_tokens=True, only_head_dropout=True
    ),
    "batchbald_samples_head": partial(
        batchbald, select_by_number_of_tokens=False, only_head_dropout=True
    ),
    # CV strategies
    "ddu_cv": ddu_sampling_cv,
}

SAMPLING_STRATEGIES = {
    "ups": ups_subsampling,
    "random": random_subsampling,
    "naive": naive_subsampling,
}


def construct_active_learner(
    model, config, initial_data, log_dir: str or Path, original_pool_size=None
):

    # TODO: rewrite using `split_by_tokens` as `strategy_kwargs`
    initial_data_copy = deepcopy(initial_data)
    use_subsampling = config.sampling_type is not None
    postfix = ""
    if ("split_by_tokens" in config) and (config.split_by_tokens):
        postfix += "_tokens"
    elif "split_by_tokens" in config:  # avoid adding "_samples" for classification
        postfix += "_samples"

    if config.strategy == "actune":
        postfix += getattr(config, "base_strategy", "_lc")

    if config.strategy == "small-text":
        postfix += getattr(config, "small_text_strategy", "_lc")

    if config.strategy == "modal":
        postfix += getattr(config, "modal_strategy", "_lc")

    query_strategy = QUERY_STRATEGIES.get(f"{config.strategy}{postfix}")
    # In this case, we assume that `config.strategy` refers to the path of the file with the strategy
    if query_strategy is None:
        query_strategy = _get_strategy_from_path(config.strategy)
    sampling_strategy = SAMPLING_STRATEGIES.get(config.sampling_type)
    if use_subsampling and sampling_strategy is None:
        sampling_strategy = _get_strategy_from_path(config.sampling_type)
    sampling_kwargs = {
        "gamma_or_k_confident_to_save": config.gamma_or_k_confident_to_save,
        "T": config.T,
    }
    strategy_kwargs = config.strategy_kwargs

    learner = ActiveLearner(
        estimator=model,
        query_strategy=query_strategy,
        train_data=initial_data_copy,
        strategy_kwargs=strategy_kwargs,
        sampling_strategy=sampling_strategy,
        sampling_kwargs=sampling_kwargs,
        log_dir=log_dir,
        original_pool_size=original_pool_size,
    )

    return learner


def _get_strategy_from_path(strategy_path):
    if (not strategy_path.endswith("/")) and ("/" in strategy_path):
        path = strategy_path[: strategy_path.rindex("/")]
    elif "/" in strategy_path:
        path = strategy_path[: strategy_path[:-2].rindex("/")]
    else:
        path = (
            Path("/".join(os.path.abspath(__file__).split("/")[:-2]))
            / "query_strategies"
        )
    strategy_name = strategy_path.split("/")[-1] or strategy_path.split("/")[-2]
    if strategy_name.endswith(".py"):
        strategy_name = strategy_name[:-3]
    sys.path.append(path)

    # For the custom function
    if "/" in strategy_path:
        strategy_func = getattr(import_module(strategy_name), strategy_name, None)
        if strategy_func is None:
            strategy_func = getattr(
                import_module(strategy_name), strategy_name + "_sampling"
            )
    else:
        package = "acleto.al4nlp.query_strategies"
        # From the library
        try:
            strategy_func = getattr(
                import_module("." + strategy_name, package=package), strategy_name, None
            )
        except:
            strategy_func = getattr(
                import_module("." + strategy_name + "_sampling", package=package),
                strategy_name + "_sampling",
                None,
            )
    return strategy_func



from ipywidgets import Layout, Label, Textarea


class TextAreaVisualizer(object):
    """Visualizer for texts via text areas."""

    def __init__(self, text_columns, width="90%", height="150px"):
        """
        Args:
            text_columns (list): List of labels (index elements) of columns that should be
                visualized via textarea.
            width (str): width of the text area.
            height (str): height of the text are.
        """
        super().__init__()
        self._text_columns = text_columns
        self._text_layout = Layout(width=width, height=height)

    def init(self, dataframe, answers):
        self._dataframe = dataframe
        self._answers = answers

    def __call__(self, index):
        """Invokes the visuzlizer.

        Args:
            dataframe (pandas.DataFrame): the dataframe that contains the data for visualization.
            index (int): the positional (iloc) index of the row to visualize.

        Returns:
            list: widgets that visualize the row
        """
        result = []
        row = self._dataframe.iloc[index]
        for label in self._text_columns:
            result.append(Label("{}:".format(label)))
            result.append(
                Textarea(value=str(row[label]), layout=self._text_layout, disabled=True)
            )

        return tuple(result)



"""
Code from https://github.com/forest-snow/alps
"""
import numpy as np
import torch
from scipy.spatial.distance import cdist

# for A100 we can use cuml
from sklearn.cluster import KMeans
from tqdm import tqdm


def closest_center_dist(X, centers):
    # return distance to closest center
    dist = torch.cdist(X, X[centers])
    cd = dist.min(axis=1).values
    return cd


def kmeans_pp(X, k, centers, **kwargs):
    # kmeans++ algorithm
    if len(centers) == 0:
        # randomly choose first center
        c1 = np.random.choice(X.size(0))
        centers.append(c1)
        k -= 1
    # greedily choose centers
    for i in tqdm(range(k)):
        dist = closest_center_dist(X, centers) ** 2
        prob = (dist / dist.sum()).cpu().detach().numpy()
        ci = np.random.choice(X.size(0), p=prob)
        centers.append(ci)
    return centers


def kmeans(X, k, tol=1e-4, **kwargs):
    # kmeans algorithm
    print("Running Kmeans")
    kmeans = KMeans(n_clusters=k).fit(X)
    centers = kmeans.cluster_centers_
    # find closest point to centers
    centroids = cdist(centers, X).argmin(axis=1)
    centroids_set = np.unique(centroids)
    m = k - len(centroids_set)
    if m > 0:
        pool = np.delete(np.arange(len(X)), centroids_set)
        p = np.random.choice(len(pool), m)
        centroids = np.concatenate((centroids_set, pool[p]), axis=None)
    return centroids


def stopping(lamda, centers):
    k = len(centers)
    return 16 * lamda * k * (np.log2(k) + 2)


def kcenter(X, k, centers, **kwargs):
    if len(centers) == 0:
        # randomly choose first center
        c1 = np.random.choice(X.size(0))
        centers.append(c1)
        k -= 1
    # greedily choose other centers
    for i in tqdm(range(k)):
        dist = closest_center_dist(X, centers)
        ci = dist.argmax().item()
        centers.append(ci)
    return centers


def badge(grads, k, **kwargs):
    # BADGE algorithm (Ash et al. 2020)
    centers = kmeans_pp(grads, k, [])
    return centers



from typing import Tuple

import numpy as np
import torch
from tqdm.auto import tqdm

from .compute_entropy import compute_conditional_entropy
from .joint_entropy import DynamicJointEntropy


def get_batchbald_batch(
    log_probs_N_K_C: torch.Tensor,
    batch_size: int,
    max_num_samples: int,
    dtype=None,
    device=None,
) -> Tuple[np.ndarray, np.ndarray]:
    """

    :param log_probs_N_K_C:
    :param batch_size:
    :param max_num_samples: how many queried samples to use to calculate the joint entropy
    :param dtype:
    :param device:
    :return:
    """
    N, K, C = log_probs_N_K_C.shape

    batch_size = min(batch_size, N)

    candidate_indices = []
    candidate_scores = []

    conditional_entropies_N = compute_conditional_entropy(log_probs_N_K_C)

    batch_joint_entropy = DynamicJointEntropy(
        max_num_samples, batch_size - 1, K, C, dtype=dtype, device=device
    )

    # We always keep these on the CPU.
    scores_N = torch.empty(N, dtype=torch.double, pin_memory=torch.cuda.is_available())

    for i in tqdm(range(batch_size), desc="BatchBALD", leave=False):
        if i > 0:
            latest_index = candidate_indices[-1]
            batch_joint_entropy.add_variables(
                log_probs_N_K_C[latest_index : latest_index + 1]
            )

        shared_conditinal_entropies = conditional_entropies_N[candidate_indices].sum()

        batch_joint_entropy.compute_batch(log_probs_N_K_C, output_entropies_B=scores_N)

        scores_N -= conditional_entropies_N + shared_conditinal_entropies
        scores_N[candidate_indices] = -float("inf")

        candidate_score, candidate_index = scores_N.max(dim=0)

        candidate_indices.append(candidate_index.item())
        candidate_scores.append(candidate_score.item())

    final_iter_scores = scores_N.numpy()
    final_iter_scores[candidate_indices] = candidate_scores

    return np.array(candidate_indices), final_iter_scores
